#!/usr/bin/env python
# coding=utf-8

# aeneas is a Python/C library and a set of tools
# to automagically synchronize audio and text (aka forced alignment)
#
# Copyright (C) 2012-2013, Alberto Pettarin (www.albertopettarin.it)
# Copyright (C) 2013-2015, ReadBeyond Srl   (www.readbeyond.it)
# Copyright (C) 2015-2017, Alberto Pettarin (www.albertopettarin.it)
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
This module contains the following classes:

* :class:`~aeneas.runtimeconfiguration.RuntimeConfiguration`,
  representing the runtime configuration.

.. versionadded:: 1.4.1
"""

from __future__ import absolute_import
from __future__ import print_function

from aeneas.configuration import Configuration
from aeneas.exacttiming import TimeValue


class RuntimeConfiguration(Configuration):
    """
    A structure representing a runtime configuration, that is,
    a set of parameters for the algorithms which process jobs and tasks.

    Allowed keys are listed below as class members.

    :param string config_string: the configuration string
    :raises: TypeError: if ``config_string`` is not ``None`` and
                        it is not a Unicode string
    :raises: KeyError: if trying to access a key not listed above
    """

    ABA_NONSPEECH_TOLERANCE = "aba_nonspeech_tolerance"
    """
    Tolerance, in seconds, for considering a given time value
    inside a nonspeech interval.

    Default: ``0.080`` seconds.

    .. versionadded:: 1.7.0
    """

    ABA_NO_ZERO_DURATION = "aba_no_zero_duration"
    """
    Offset, in seconds, to be added to fragments with zero length.

    Default: ``0.001`` seconds.

    .. versionadded:: 1.7.0
    """

    ALLOW_UNLISTED_LANGUAGES = "allow_unlisted_languages"
    """
    If ``True``, allow using a language code
    not listed in the TTS supported languages list;
    otherwise, generate an error if the user attempts
    to use a language not listed.

    Default: ``False``.

    .. versionadded:: 1.4.1
    """

    C_EXTENSIONS = "c_extensions"
    """
    If ``True`` and the Python C/C++ extensions
    are available, use them.
    Otherwise, use the pure Python code.

    This option is equivalent to
    setting ``CDTW``, ``CEW``, ``CFW``,
    and ``CMFCC`` to ``True`` or ``False`` at once.

    Default: ``True``.

    .. versionadded:: 1.4.1
    """

    CDTW = "cdtw"
    """
    If ``True`` and the Python C extension ``cdtw``
    is available, use it.
    Otherwise, use the pure Python code.

    Default: ``True``.

    .. versionadded:: 1.5.1
    """

    CEW = "cew"
    """
    If ``True`` and the Python C extension ``cew``
    is available, use it.
    Otherwise, use the pure Python code.

    Default: ``True``.

    .. versionadded:: 1.5.1
    """

    CFW = "cfw"
    """
    If ``True`` and the Python C++ extension ``cfw``
    is available, use it.
    Otherwise, use the pure Python code.

    Default: ``True``.

    .. versionadded:: 1.6.0
    """

    CEW_SUBPROCESS_ENABLED = "cew_subprocess_enabled"
    """
    If ``True``, calls to ``aeneas.cew``
    will be done via ``subprocess``, using the
    :class:`~aeneas.cewsubprocess.CEWSubprocess`
    helper class.

    Default: ``False``.

    .. versionadded:: 1.5.0
    """

    CEW_SUBPROCESS_PATH = "cew_subprocess_path"
    """
    Use the given path to the python executable
    when calling ``aeneas.cew`` via ``subprocess``.

    You might need to use a full path, like ``/path/to/your/python``.

    Default: ``python``.

    .. versionadded:: 1.5.0
    """

    CMFCC = "cmfcc"
    """
    If ``True`` and the Python C extension ``cmfcc``
    is available, use it.
    Otherwise, use the pure Python code.

    Default: ``True``.

    .. versionadded:: 1.5.1
    """

    DOWNLOADER_SLEEP = "downloader_sleep"
    """
    Wait this number of seconds before the next HTTP POST request
    of the ``Downloader``.
    This parameter can be used to throttle the HTTP usage.
    It cannot be a negative value.

    Default: ``1.000``.

    .. versionadded:: 1.7.2
    """

    DOWNLOADER_RETRY_ATTEMPTS = "downloader_retry_attempts"
    """
    Retry an HTTP POST request generated by the ``Downloader``
    for this number of times before giving up.
    It must be an integer greater than zero.

    Default: ``5``.

    .. versionadded:: 1.7.2
    """

    DTW_ALGORITHM = "dtw_algorithm"
    """
    DTW aligner algorithm.

    Allowed values:

    * :data:`~aeneas.dtw.DTWAlgorithm.EXACT` (``exact``)
    * :data:`~aeneas.dtw.DTWAlgorithm.STRIPE` (``stripe``, default)

    .. versionadded:: 1.4.1
    """

    DTW_MARGIN = "dtw_margin"
    """
    DTW aligner margin, in seconds, for the ``stripe`` algorithm.

    Default: ``60``, corresponding to ``60 s`` ahead and behind
    (i.e., ``120 s`` total margin).

    .. versionadded:: 1.4.1
    """

    DTW_MARGIN_L1 = "dtw_margin_l1"
    """
    DTW aligner margin, in seconds, for the ``stripe`` algorithm
    at level 1 (paragraph).

    Default: ``60``, corresponding to ``60 s`` ahead and behind
    (i.e., ``120 s`` total margin).

    .. versionadded:: 1.7.0
    """

    DTW_MARGIN_L2 = "dtw_margin_l2"
    """
    DTW aligner margin, in seconds, for the ``stripe`` algorithm
    at level 2 (sentence).

    Default: ``30``, corresponding to ``30 s`` ahead and behind
    (i.e., ``60 s`` total margin).

    .. versionadded:: 1.7.0
    """

    DTW_MARGIN_L3 = "dtw_margin_l3"
    """
    DTW aligner margin, in seconds, for the ``stripe`` algorithm
    at level 3 (word).

    Default: ``10``, corresponding to ``10 s`` ahead and behind
    (i.e., ``20s`` total margin).

    .. versionadded:: 1.7.0
    """

    FFMPEG_PATH = "ffmpeg_path"
    """
    Path to the ``ffmpeg`` executable.

    You might need to use a full path, like ``/path/to/your/ffmpeg``.

    Default: ``ffmpeg``.

    .. versionadded:: 1.4.1
    """

    FFMPEG_SAMPLE_RATE = "ffmpeg_sample_rate"
    """
    Sample rate for ``ffmpeg``, in Hertz.

    Default: ``16000``.

    .. versionadded:: 1.4.1
    """

    FFPROBE_PATH = "ffprobe_path"
    """
    Path to the ``ffprobe`` executable.

    You might use a full path, like ``/path/to/your/ffprobe``.

    Default: ``ffprobe``.

    .. versionadded:: 1.4.1
    """

    JOB_MAX_TASKS = "job_max_tasks"
    """
    Maximum number of Tasks of a Job.
    If a Job has more Tasks than this value,
    it will not be executed and an error will be raised.
    Use ``0`` for disabling this check.

    Default: ``0`` (disabled).

    .. versionadded:: 1.4.1
    """

    MFCC_FILTERS = "mfcc_filters"
    """
    Number of filters for extracting MFCCs.

    Default: ``40``.

    .. versionadded:: 1.4.1
    """

    MFCC_SIZE = "mfcc_size"
    """
    Number of MFCCs to extract, including the 0th.

    Default: ``13``.

    .. versionadded:: 1.4.1
    """

    MFCC_FFT_ORDER = "mfcc_fft_order"
    """
    Order of the RFFT for extracting MFCCs.
    It must be a power of two.

    Default: ``512``.

    .. versionadded:: 1.4.1
    """

    MFCC_LOWER_FREQUENCY = "mfcc_lower_frequency"
    """
    Lower frequency to be used for extracting MFCCs, in Hertz.

    Default: ``133.3333``.

    .. versionadded:: 1.4.1
    """

    MFCC_UPPER_FREQUENCY = "mfcc_upper_frequency"
    """
    Upper frequency to be used for extracting MFCCs, in Hertz.

    Default: ``6855.4976``.

    .. versionadded:: 1.4.1
    """

    MFCC_EMPHASIS_FACTOR = "mfcc_emphasis_factor"
    """
    Emphasis factor to be applied to MFCCs.

    Default: ``0.970``.

    .. versionadded:: 1.4.1
    """

    MFCC_MASK_NONSPEECH = "mfcc_mask_nonspeech"
    """
    If ``True``, computes the DTW path ignoring nonspeech frames.
    Setting this parameter to ``True`` might help aligning
    at word level granularity.

    Default: ``False``.

    .. versionadded:: 1.7.0
    """

    MFCC_WINDOW_LENGTH = "mfcc_window_length"
    """
    Length of the window for extracting MFCCs, in seconds.
    It is usual to set it between 1.5 and 4 times
    the value of ``MFCC_WINDOW_SHIFT``.

    Default: ``0.100``.

    .. versionadded:: 1.4.1
    """

    MFCC_WINDOW_SHIFT = "mfcc_window_shift"
    """
    Shift of the window for extracting MFCCs, in seconds.
    This parameter is basically the time step
    of the synchronization maps output.

    Default: ``0.040``.

    .. versionadded:: 1.4.1
    """

    MFCC_MASK_NONSPEECH_L1 = "mfcc_mask_nonspeech_l1"
    """
    If ``True``, computes the DTW path ignoring nonspeech frames
    at level 1 (paragraph).

    Default: ``False``.

    .. versionadded:: 1.7.0
    """

    MFCC_WINDOW_LENGTH_L1 = "mfcc_window_length_l1"
    """
    Length of the window, in seconds,
    for extracting MFCCs at level 1 (paragraph).
    It is usual to set it between 1.5 and 4 times
    the value of ``MFCC_WINDOW_SHIFT_L1``.

    Default: ``0.100``.

    .. versionadded:: 1.5.0
    """

    MFCC_WINDOW_SHIFT_L1 = "mfcc_window_shift_l1"
    """
    Shift of the window, in seconds,
    for extracting MFCCs at level 1 (paragraph).
    This parameter is basically the time step
    of the synchronization map output at level 1.

    Default: ``0.040``.

    .. versionadded:: 1.5.0
    """

    MFCC_MASK_NONSPEECH_L2 = "mfcc_mask_nonspeech_l2"
    """
    If ``True``, computes the DTW path ignoring nonspeech frames
    at level 2 (sentence).

    Default: ``False``.

    .. versionadded:: 1.7.0
    """

    MFCC_WINDOW_LENGTH_L2 = "mfcc_window_length_l2"
    """
    Length of the window, in seconds,
    for extracting MFCCs at level 2 (sentence).
    It is usual to set it between 1.5 and 4 times
    the value of ``MFCC_WINDOW_SHIFT_L2``.

    Default: ``0.050``.

    .. versionadded:: 1.5.0
    """

    MFCC_WINDOW_SHIFT_L2 = "mfcc_window_shift_l2"
    """
    Shift of the window, in seconds,
    for extracting MFCCs at level 2 (sentence).
    This parameter is basically the time step
    of the synchronization map output at level 2.

    Default: ``0.020``.

    .. versionadded:: 1.5.0
    """

    MFCC_MASK_NONSPEECH_L3 = "mfcc_mask_nonspeech_l3"
    """
    If ``True``, computes the DTW path ignoring nonspeech frames
    at level 3 (word).

    Default: ``False``.

    .. versionadded:: 1.7.0
    """

    MFCC_WINDOW_LENGTH_L3 = "mfcc_window_length_l3"
    """
    Length of the window, in seconds,
    for extracting MFCCs at level 3 (word).
    It is usual to set it between 1.5 and 4 times
    the value of ``MFCC_WINDOW_SHIFT_L3``.

    Default: ``0.020``.

    .. versionadded:: 1.5.0
    """

    MFCC_WINDOW_SHIFT_L3 = "mfcc_window_shift_l3"
    """
    Shift of the window, in seconds,
    for extracting MFCCs at level 3 (word).
    This parameter is basically the time step
    of the synchronization map output at level 3.

    Default: ``0.005``.

    .. versionadded:: 1.5.0
    """

    MFCC_GRANULARITY_MAP = {
        1: (
            DTW_MARGIN_L1,
            MFCC_MASK_NONSPEECH_L1,
            MFCC_WINDOW_LENGTH_L1,
            MFCC_WINDOW_SHIFT_L1
        ),
        2: (
            DTW_MARGIN_L2,
            MFCC_MASK_NONSPEECH_L2,
            MFCC_WINDOW_LENGTH_L2,
            MFCC_WINDOW_SHIFT_L2
        ),
        3: (
            DTW_MARGIN_L3,
            MFCC_MASK_NONSPEECH_L3,
            MFCC_WINDOW_LENGTH_L3,
            MFCC_WINDOW_SHIFT_L3
        ),
    }
    """
    Map level numbers to
    ``DTW_MARGIN_*``,
    ``MFCC_MASK_NONSPEECH_*``,
    ``MFCC_WINDOW_LENGTH_*``,
    and ``MFCC_WINDOW_SHIFT_*``
    keys.

    .. versionadded:: 1.5.0
    """

    MFCC_MASK_EXTEND_SPEECH_INTERVAL_AFTER = "mfcc_mask_extend_speech_after"
    """
    Extend to the right (after/future)
    a speech interval found by the VAD algorithm,
    by this many frames, when masking nonspeech out.

    Default: ``0``.

    .. versionadded:: 1.7.0
    """

    MFCC_MASK_EXTEND_SPEECH_INTERVAL_BEFORE = "mfcc_mask_extend_speech_before"
    """
    Extend to the left (before/past)
    a speech interval found by the VAD algorithm,
    by this many frames, when masking nonspeech out.

    Default: ``0``.

    .. versionadded:: 1.7.0
    """

    MFCC_MASK_LOG_ENERGY_THRESHOLD = "mfcc_mask_log_energy_threshold"
    """
    Threshold for the VAD algorithm to decide
    that a given frame contains speech, when masking nonspeech out.
    Note that this is the log10 of the energy coefficient.

    Default: ``0.699`` = ``log10(5)``, that is, a frame must have
    an energy at least 5 times higher than the minimum
    to be considered a speech frame.

    .. versionadded:: 1.7.0
    """

    MFCC_MASK_MIN_NONSPEECH_LENGTH = "mfcc_mask_min_nonspeech_length"
    """
    Minimum length, in frames, of a nonspeech interval to be masked out.

    Default: ``1``.

    .. versionadded:: 1.7.0
    """

    NUANCE_TTS_API_ID = "nuance_tts_api_id"
    """
    Your ID value to use the Nuance TTS API.

    You will be billed according to your Nuance Developers account plan.

    Important: this feature is experimental, use at your own risk.
    It is recommended not to use this TTS at word-level granularity,
    as it will create many requests, hence it will be expensive.
    If you still want to use it, you can enable
    the TTS caching mechanism by setting
    :data:`~aeneas.runtimeconfiguration.RuntimeConfiguration.TTS_CACHE`
    to ``True``.

    .. versionadded:: 1.5.0
    """

    NUANCE_TTS_API_KEY = "nuance_tts_api_key"
    """
    Your KEY value to use the Nuance TTS API.

    You will be billed according to your Nuance Developers account plan.

    Important: this feature is experimental, use at your own risk.
    It is recommended not to use this TTS at word-level granularity,
    as it will create many requests, hence it will be expensive.
    If you still want to use it, you can enable
    the TTS caching mechanism by setting
    :data:`~aeneas.runtimeconfiguration.RuntimeConfiguration.TTS_CACHE`
    to ``True``.

    .. versionadded:: 1.5.0
    """

    SAFETY_CHECKS = "safety_checks"
    """
    If ``True``, perform safety checks on input files and parameters.
    If set to ``False``, it disables:

    * checks perfomed by :class:`~aeneas.validator.Validator`;
    * converting the audio file synthesized by the TTS engine so that its sample rate times the MFCC shift is an integer value.

    .. warning:: Setting this parameter to ``False`` might result in runtime errors. Please be sure to understand the implications.

    Default: ``True``.

    .. versionadded:: 1.7.0
    """

    TASK_MAX_AUDIO_LENGTH = "task_max_audio_length"
    """
    Maximum length of the audio file of a Task, in seconds.
    If a Task has an audio file longer than this value,
    it will not be executed and an error will be raised.

    Use ``0`` to disable this check.

    Default: ``0`` seconds.

    .. versionadded:: 1.4.1
    """

    TASK_MAX_TEXT_LENGTH = "task_max_text_length"
    """
    Maximum number of text fragments in the text file of a Task.
    If a Task has more text fragments than this value,
    it will not be executed and an error will be raised.

    Use ``0`` to disable this check.

    Default: ``0`` (disabled).

    .. versionadded:: 1.4.1
    """

    TMP_PATH = "tmp_path"
    """
    Path to the temporary directory to be used.
    Default: ``None``, meaning that the default temporary directory
    will be set by
    :data:`~aeneas.globalconstants.TMP_PATH_DEFAULT_POSIX`
    or
    :data:`~aeneas.globalconstants.TMP_PATH_DEFAULT_NONPOSIX`
    depending on your OS.

    .. versionadded:: 1.4.1
    """

    TTS = "tts"
    """
    The TTS engine to use for synthesizing text.

    Allowed values are listed in :data:`~aeneas.synthesizer.Synthesizer.ALLOWED_VALUES`.

    The default value is
    :data:`~aeneas.synthesizer.Synthesizer.ESPEAK` (``espeak``)
    which will use the built-in eSpeak TTS wrapper.
    You might need to provide a ``/full/path/to/your/espeak`` value
    to the
    :data:`~aeneas.runtimeconfiguration.RuntimeConfiguration.TTS_PATH`
    parameter if the command ``espeak`` is not available in
    one of the directories listed in your ``PATH`` environment variable.

    Specify the value
    :data:`~aeneas.synthesizer.Synthesizer.ESPEAKNG` (``espeak-ng``)
    to use the eSpeak-ng TTS wrapper.
    You might need to provide a ``/full/path/to/your/espeak-ng`` value
    to the
    :data:`~aeneas.runtimeconfiguration.RuntimeConfiguration.TTS_PATH`
    parameter if the command ``espeak-ng`` is not available in
    one of the directories listed in your ``PATH`` environment variable.

    Specify the value
    :data:`~aeneas.synthesizer.Synthesizer.FESTIVAL` (``festival``)
    to use the built-in Festival TTS wrapper.
    You might need to provide a ``/full/path/to/your/text2wave`` value
    to the
    :data:`~aeneas.runtimeconfiguration.RuntimeConfiguration.TTS_PATH`
    parameter if the command ``text2wave`` is not available in
    one of the directories listed in your ``PATH`` environment variable.

    Specify the value
    :data:`~aeneas.synthesizer.Synthesizer.AWS` (``aws``)
    to use the built-in AWS Polly TTS API wrapper;
    you will need to provide your AWS API Access Key and Secret Access Key
    by either storing them on disk
    (e.g., in ``~/.aws/credentials`` and ``~/.aws/config``)
    or setting them in environment variables.
    Please refer to
    http://boto3.readthedocs.io/en/latest/guide/configuration.html
    for further details.

    Specify the value
    :data:`~aeneas.synthesizer.Synthesizer.NUANCE` (``nuance``)
    to use the built-in Nuance TTS API wrapper;
    you will need to provide your Nuance Developer API ID and API Key using the
    :data:`~aeneas.runtimeconfiguration.RuntimeConfiguration.NUANCE_TTS_API_ID`
    and
    :data:`~aeneas.runtimeconfiguration.RuntimeConfiguration.NUANCE_TTS_API_KEY`
    parameters.
    Please note that you will be billed according to your Nuance Developers account plan.

    Specify the value
    :data:`~aeneas.synthesizer.Synthesizer.CUSTOM` (``custom``)
    to use a custom TTS;
    you will need to provide the path to the Python source file
    containing your TTS wrapper using the
    :data:`~aeneas.runtimeconfiguration.RuntimeConfiguration.TTS_PATH`
    parameter.

    .. versionadded:: 1.5.0
    """

    TTS_PATH = "tts_path"
    """
    Path to the TTS engine executable
    or the Python CustomTTSWrapper ``.py`` source file
    (see the ``aeneas/extra`` directory for examples).

    You might need to use a full path,
    like ``/path/to/your/ttsengine`` or
    ``/path/to/your/ttswrapper.py``.

    Default: ``None``, implying to use the default path
    defined by each TTS wrapper, if it calls the TTS engine
    via ``subprocess`` (otherwise it does not matter).

    .. versionadded:: 1.5.0
    """

    TTS_VOICE_CODE = "tts_voice_code"
    """
    The code of the TTS voice to use.
    If you specify this value, it will override the default voice code
    associated with the language of your text.

    Default: ``None``.

    .. versionadded:: 1.5.0
    """

    TTS_CACHE = "tts_cache"
    """
    If set to ``True``, synthesize each distinct text fragment
    only once, caching the resulting audio data as a file on disk.

    The cache files will be removed after the synthesis is compled.

    This option is useful when calling TTS engines,
    via subprocess or remote APIs,
    on text files with many identical fragments,
    for example when aligning at word-level granularity.

    Enabling this option will create the cache files in
    :data:`~aeneas.runtimeconfiguration.RuntimeConfiguration.TMP_PATH`,
    so make sure that that path has enough free space.

    Default: ``False``.

    .. versionadded:: 1.6.0
    """

    TTS_API_SLEEP = "tts_api_sleep"
    """
    Wait this number of seconds before the next HTTP POST request
    to the Nuance TTS API.
    This parameter can be used to throttle the HTTP usage.
    It cannot be a negative value.

    Note that this parameter was called ``nuance_tts_api_sleep``
    before v1.7.0.

    Default: ``1.000``.

    .. versionadded:: 1.5.0
    """

    TTS_API_RETRY_ATTEMPTS = "tts_api_retry_attempts"
    """
    Retry an HTTP POST request to the Nuance TTS API
    for this number of times before giving up.
    It must be an integer greater than zero.

    Note that this parameter was called ``nuance_tts_api_retry_attempts``
    before v1.7.0.

    Default: ``5``.

    .. versionadded:: 1.5.0
    """

    TTS_L1 = "tts_l1"
    """
    The TTS engine to use for synthesizing text
    at level 1 (paragraph).

    See also :data:`~aeneas.runtimeconfiguration.RuntimeConfiguration.TTS`.

    Default: ``espeak``.

    .. versionadded:: 1.6.0
    """

    TTS_PATH_L1 = "tts_path_l1"
    """
    Path to the TTS engine executable to use for synthesizing text
    at level 1 (paragraph).

    See also :data:`~aeneas.runtimeconfiguration.RuntimeConfiguration.TTS_PATH`.

    Default: ``None``.

    .. versionadded:: 1.6.0
    """

    TTS_L2 = "tts_l2"
    """
    The TTS engine to use for synthesizing text
    at level 2 (sentence).

    See also :data:`~aeneas.runtimeconfiguration.RuntimeConfiguration.TTS`.

    Default: ``espeak``.

    .. versionadded:: 1.6.0
    """

    TTS_PATH_L2 = "tts_path_l2"
    """
    Path to the TTS engine executable to use for synthesizing text
    at level 2 (sentence).

    See also :data:`~aeneas.runtimeconfiguration.RuntimeConfiguration.TTS_PATH`.

    Default: ``None``.

    .. versionadded:: 1.6.0
    """

    TTS_L3 = "tts_l3"
    """
    The TTS engine to use for synthesizing text
    at level 3 (word).

    See also :data:`~aeneas.runtimeconfiguration.RuntimeConfiguration.TTS`.

    Default: ``espeak``.

    .. versionadded:: 1.6.0
    """

    TTS_PATH_L3 = "tts_path_l3"
    """
    Path to the TTS engine executable to use for synthesizing text
    at level 3 (word).

    See also :data:`~aeneas.runtimeconfiguration.RuntimeConfiguration.TTS_PATH`.

    Default: ``None``.

    .. versionadded:: 1.6.0
    """

    TTS_GRANULARITY_MAP = {
        1: (TTS_L1, TTS_PATH_L1),
        2: (TTS_L2, TTS_PATH_L2),
        3: (TTS_L3, TTS_PATH_L3),
    }
    """
    Map level numbers to ``TTS_*``
    and ``TTS_PATH_*`` keys.

    .. versionadded:: 1.6.0
    """

    VAD_EXTEND_SPEECH_INTERVAL_AFTER = "vad_extend_speech_after"
    """
    Extend to the right (after/future)
    a speech interval found by the VAD algorithm,
    by this many seconds.

    Default: ``0`` seconds.

    .. versionadded:: 1.4.1
    """

    VAD_EXTEND_SPEECH_INTERVAL_BEFORE = "vad_extend_speech_before"
    """
    Extend to the left (before/past)
    a speech interval found by the VAD algorithm,
    by this many seconds.

    Default: ``0`` seconds.

    .. versionadded:: 1.4.1
    """

    VAD_LOG_ENERGY_THRESHOLD = "vad_log_energy_threshold"
    """
    Threshold for the VAD algorithm to decide
    that a given frame contains speech.
    Note that this is the log10 of the energy coefficient.

    Default: ``0.699`` = ``log10(5)``, that is, a frame must have
    an energy at least 5 times higher than the minimum
    to be considered a speech frame.

    .. versionadded:: 1.4.1
    """

    VAD_MIN_NONSPEECH_LENGTH = "vad_min_nonspeech_length"
    """
    Minimum length, in seconds, of a nonspeech interval.

    Default: ``0.200`` seconds.

    .. versionadded:: 1.4.1
    """

    # NOTE not using aliases just not to become confused
    #      about external (user rconf) and internal (lib code) key names
    #      although the functionality might be useful in the future
    FIELDS = [
        (ABA_NONSPEECH_TOLERANCE, ("0.080", TimeValue, [], u"adjust nonspeech tolerance, in s")),
        (ABA_NO_ZERO_DURATION, ("0.001", TimeValue, [], u"add this shift to zero length fragments, in s")),
        (ALLOW_UNLISTED_LANGUAGES, (False, bool, [], u"if True, allow languages not listed")),

        (C_EXTENSIONS, (True, bool, [], u"run C/C++ extensions")),
        (CDTW, (True, bool, [], u"run C extension cdtw")),
        (CEW, (True, bool, [], u"run C extension cew")),
        (CFW, (True, bool, [], u"run C++ extension cfw")),
        (CMFCC, (True, bool, [], u"run C extension cmfcc")),

        (CEW_SUBPROCESS_ENABLED, (False, bool, [], u"run cew in separate subprocess")),
        (CEW_SUBPROCESS_PATH, ("python", None, [], u"path to python executable")),          # or a full path like "/usr/bin/python"

        (DTW_ALGORITHM, ("stripe", None, [], u"DTW algorithm (stripe, exact)")),
        (DTW_MARGIN, ("60.000", TimeValue, [], u"DTW margin, in s")),

        (DOWNLOADER_SLEEP, ("1.000", TimeValue, [], u"sleep between Downloader calls, in s")),
        (DOWNLOADER_RETRY_ATTEMPTS, (5, int, [], u"number of retries for a failed Downloader call")),

        (FFMPEG_PATH, ("ffmpeg", None, [], u"path to ffmpeg executable")),                  # or a full path like "/usr/bin/ffmpeg"
        (FFMPEG_SAMPLE_RATE, (16000, int, [], u"ffmpeg sample rate")),

        (FFPROBE_PATH, ("ffprobe", None, [], u"path to ffprobe executable")),               # or a full path like "/usr/bin/ffprobe"

        (JOB_MAX_TASKS, (0, int, [], u"max number of tasks per job (0 to disable)")),

        (MFCC_FILTERS, (40, int, [], u"number of MFCC filters")),
        (MFCC_SIZE, (13, int, [], u"number of MFCC")),
        (MFCC_FFT_ORDER, (512, int, [], u"FFT order for computing MFCC")),
        (MFCC_LOWER_FREQUENCY, (133.3333, float, [], u"MFCC lower frequency cutoff, in Hz")),
        (MFCC_UPPER_FREQUENCY, (6855.4976, float, [], u"MFCC upper frequency cutoff, in Hz")),
        (MFCC_EMPHASIS_FACTOR, (0.970, float, [], u"MFCC emphasis factor")),

        (MFCC_MASK_NONSPEECH, (False, bool, [], u"if True, mask MFCC nonspeech frames")),
        (MFCC_WINDOW_LENGTH, ("0.100", TimeValue, [], u"MFCC window length, in s")),
        (MFCC_WINDOW_SHIFT, ("0.040", TimeValue, [], u"MFCC window shift, in s")),

        (MFCC_MASK_EXTEND_SPEECH_INTERVAL_AFTER, (0, int, [], u"when masking MFCC, extend speech interval after, in frames")),
        (MFCC_MASK_EXTEND_SPEECH_INTERVAL_BEFORE, (0, int, [], u"when masking MFCC, extend speech interval before, in frames")),
        (MFCC_MASK_LOG_ENERGY_THRESHOLD, (0.699, float, [], u"when masking MFCC, log energy threshold for speech")),
        (MFCC_MASK_MIN_NONSPEECH_LENGTH, (1, int, [], u"when masking MFCC, min nonspeech interval length, in frames")),

        (DTW_MARGIN_L1, ("60.000", TimeValue, [], u"level 1 (para) DTW margin, in s")),
        (MFCC_MASK_NONSPEECH_L1, (False, bool, [], u"if True, mask MFCC nonspeech frames on level 1 (para)")),
        (MFCC_WINDOW_LENGTH_L1, ("0.100", TimeValue, [], u"level 1 (para) MFCC window length, in s")),
        (MFCC_WINDOW_SHIFT_L1, ("0.040", TimeValue, [], u"level 1 (para) MFCC window shift, in s")),
        (DTW_MARGIN_L2, ("30.000", TimeValue, [], u"level 2 (sent) DTW margin, in s")),
        (MFCC_MASK_NONSPEECH_L2, (False, bool, [], u"if True, mask MFCC nonspeech frames on level 2 (sent)")),
        (MFCC_WINDOW_LENGTH_L2, ("0.050", TimeValue, [], u"level 2 (sent) MFCC window length, in s")),
        (MFCC_WINDOW_SHIFT_L2, ("0.020", TimeValue, [], u"level 2 (sent) MFCC window shift, in s")),
        (DTW_MARGIN_L3, ("10.000", TimeValue, [], u"level 3 (word) DTW margin, in s")),
        (MFCC_MASK_NONSPEECH_L3, (False, bool, [], u"if True, mask MFCC nonspeech frames on level 3 (word)")),
        (MFCC_WINDOW_LENGTH_L3, ("0.020", TimeValue, [], u"level 3 (word) MFCC window length, in s")),
        (MFCC_WINDOW_SHIFT_L3, ("0.005", TimeValue, [], u"level 3 (word) MFCC window shift, in s")),

        (NUANCE_TTS_API_ID, (None, None, [], u"Nuance Developer API ID")),
        (NUANCE_TTS_API_KEY, (None, None, [], u"Nuance Developer API Key")),

        (SAFETY_CHECKS, (True, bool, [], u"if True, always perform safety checks")),

        (TASK_MAX_AUDIO_LENGTH, ("0", TimeValue, [], u"max length of single audio file, in s (0 to disable)")),
        (TASK_MAX_TEXT_LENGTH, (0, int, [], u"max length of single text file, in fragments (0 to disable)")),

        (TMP_PATH, (None, None, [], u"path to the temporary dir")),

        (TTS, ("espeak", None, [], u"TTS wrapper to use")),
        (TTS_PATH, (None, None, [], u"path of the TTS executable/wrapper")),                # None (= default) or "espeak" or "/usr/bin/espeak"
        (TTS_VOICE_CODE, (None, None, [], u"overrides TTS voice code selected by language with this value")),
        (TTS_CACHE, (False, bool, [], u"if True, cache synthesized audio files")),
        (TTS_API_SLEEP, ("1.000", TimeValue, [], u"sleep between TTS API calls, in s")),
        (TTS_API_RETRY_ATTEMPTS, (5, int, [], u"number of retries for a failed TTS API call")),

        (TTS_L1, ("espeak", None, [], u"TTS wrapper to use at level 1 (para)")),
        (TTS_PATH_L1, (None, None, [], u"path to level 1 (para) TTS executable/wrapper")),  # None (= default) or "espeak" or "/usr/bin/espeak"
        (TTS_L2, ("espeak", None, [], u"TTS wrapper to use at level 2 (sent)")),
        (TTS_PATH_L2, (None, None, [], u"path to level 2 (sent) TTS executable/wrapper")),  # None (= default) or "espeak" or "/usr/bin/espeak"
        (TTS_L3, ("espeak", None, [], u"TTS wrapper to use at level 3 (word)")),
        (TTS_PATH_L3, (None, None, [], u"path to level 3 (word) TTS executable/wrapper")),  # None (= default) or "espeak" or "/usr/bin/espeak"

        (VAD_EXTEND_SPEECH_INTERVAL_AFTER, ("0.000", TimeValue, [], u"extend speech interval after, in s")),
        (VAD_EXTEND_SPEECH_INTERVAL_BEFORE, ("0.000", TimeValue, [], u"extend speech interval before, in s")),
        (VAD_LOG_ENERGY_THRESHOLD, (0.699, float, [], u"log energy threshold for speech")),
        (VAD_MIN_NONSPEECH_LENGTH, ("0.200", TimeValue, [], u"min nonspeech interval length, in s")),
    ]

    TAG = u"RuntimeConfiguration"

    def __init__(self, config_string=None):
        super(RuntimeConfiguration, self).__init__(config_string)

    @property
    def safety_checks(self):
        """
        Return the value of the
        :data:`~aeneas.runtimeconfiguration.RuntimeConfiguration.SAFETY_CHECKS`
        key stored in this configuration object.

        If ``False``, safety checks are not performed.

        :rtype: bool
        """
        return self[self.SAFETY_CHECKS]

    @property
    def sample_rate(self):
        """
        Return the value of the
        :data:`~aeneas.runtimeconfiguration.RuntimeConfiguration.FFMPEG_SAMPLE_RATE`
        key stored in this configuration object.

        :rtype: int
        """
        return self[self.FFMPEG_SAMPLE_RATE]

    @property
    def dtw_margin(self):
        """
        Return the value of the
        :data:`~aeneas.runtimeconfiguration.RuntimeConfiguration.DTW_MARGIN`
        key stored in this configuration object.

        :rtype: :class:`~aeneas.exacttiming.TimeValue`
        """
        return self[self.DTW_MARGIN]

    @property
    def mmn(self):
        """
        Return the value of the
        :data:`~aeneas.runtimeconfiguration.RuntimeConfiguration.MFCC_MASK_NONSPEECH`
        key stored in this configuration object.

        :rtype: bool
        """
        return self[self.MFCC_MASK_NONSPEECH]

    @property
    def mws(self):
        """
        Return the value of the
        :data:`~aeneas.runtimeconfiguration.RuntimeConfiguration.MFCC_WINDOW_SHIFT`
        key stored in this configuration object.

        :rtype: :class:`~aeneas.exacttiming.TimeValue`
        """
        return self[self.MFCC_WINDOW_SHIFT]

    @property
    def mwl(self):
        """
        Return the value of the
        :data:`~aeneas.runtimeconfiguration.RuntimeConfiguration.MFCC_WINDOW_LENGTH`
        key stored in this configuration object.

        :rtype: :class:`~aeneas.exacttiming.TimeValue`
        """
        return self[self.MFCC_WINDOW_LENGTH]

    @property
    def tts(self):
        """
        Return the value of the
        :data:`~aeneas.runtimeconfiguration.RuntimeConfiguration.TTS`
        key stored in this configuration object.

        :rtype: string
        """
        return self[self.TTS]

    @property
    def tts_path(self):
        """
        Return the value of the
        :data:`~aeneas.runtimeconfiguration.RuntimeConfiguration.TTS_PATH`
        key stored in this configuration object.

        :rtype: string
        """
        return self[self.TTS_PATH]

    def set_granularity(self, level):
        """
        Set the values for
        :data:`~aeneas.runtimeconfiguration.RuntimeConfiguration.MFCC_WINDOW_LENGTH`
        and
        :data:`~aeneas.runtimeconfiguration.RuntimeConfiguration.MFCC_WINDOW_SHIFT`
        matching the given granularity level.

        Currently supported levels:

        * ``1`` (paragraph)
        * ``2`` (sentence)
        * ``3`` (word)

        :param int level: the desired granularity level
        """
        if level in self.MFCC_GRANULARITY_MAP.keys():
            margin_key, mask_key, length_key, shift_key = self.MFCC_GRANULARITY_MAP[level]
            self[self.DTW_MARGIN] = self[margin_key]
            self[self.MFCC_MASK_NONSPEECH] = self[mask_key]
            self[self.MFCC_WINDOW_LENGTH] = self[length_key]
            self[self.MFCC_WINDOW_SHIFT] = self[shift_key]

    def set_tts(self, level):
        """
        Set the values for
        :data:`~aeneas.runtimeconfiguration.RuntimeConfiguration.TTS`
        and
        :data:`~aeneas.runtimeconfiguration.RuntimeConfiguration.TTS_PATH`
        matching the given granularity level.

        Currently supported levels:

        * ``1`` (paragraph)
        * ``2`` (sentence)
        * ``3`` (word)

        :param int level: the desired granularity level
        """
        if level in self.TTS_GRANULARITY_MAP.keys():
            tts_key, tts_path_key = self.TTS_GRANULARITY_MAP[level]
            self[self.TTS] = self[tts_key]
            self[self.TTS_PATH] = self[tts_path_key]



#!/usr/bin/env python
# coding=utf-8

# aeneas is a Python/C library and a set of tools
# to automagically synchronize audio and text (aka forced alignment)
#
# Copyright (C) 2012-2013, Alberto Pettarin (www.albertopettarin.it)
# Copyright (C) 2013-2015, ReadBeyond Srl   (www.readbeyond.it)
# Copyright (C) 2015-2017, Alberto Pettarin (www.albertopettarin.it)
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
Extract MFCCs from a given audio file.
"""

from __future__ import absolute_import
from __future__ import print_function
import io
import sys
import numpy

from aeneas.audiofile import AudioFileConverterError
from aeneas.audiofile import AudioFileNotInitializedError
from aeneas.audiofile import AudioFileUnsupportedFormatError
from aeneas.audiofilemfcc import AudioFileMFCC
from aeneas.runtimeconfiguration import RuntimeConfiguration
from aeneas.tools.abstract_cli_program import AbstractCLIProgram
import aeneas.globalfunctions as gf


class ExtractMFCCCLI(AbstractCLIProgram):
    """
    Extract MFCCs from a given audio file.
    """
    INPUT_FILE = gf.relative_path("res/audio.wav", __file__)
    OUTPUT_FILE = "output/audio.wav.mfcc.txt"

    NAME = gf.file_name_without_extension(__file__)

    HELP = {
        "description": u"Extract MFCCs from a given audio file as a fat matrix.",
        "synopsis": [
            (u"AUDIO_FILE OUTPUT_FILE", True)
        ],
        "examples": [
            u"%s %s" % (INPUT_FILE, OUTPUT_FILE)
        ],
        "options": [
            u"-b, --binary : output MFCCs as a float64 binary file",
            u"-d, --delete-first : do not output the 0th MFCC coefficient",
            u"-n, --npy : output MFCCs as a NumPy .npy binary file",
            u"-t, --transpose : transpose the MFCCs matrix, returning a tall matrix",
            u"-z, --npz : output MFCCs as a NumPy compressed .npz binary file",
            u"--format=FMT : output to text file using format FMT (default: '%.18e')"
        ]
    }

    def perform_command(self):
        """
        Perform command and return the appropriate exit code.

        :rtype: int
        """
        if len(self.actual_arguments) < 2:
            return self.print_help()
        input_file_path = self.actual_arguments[0]
        output_file_path = self.actual_arguments[1]

        output_text_format = self.has_option_with_value(u"--format")
        if output_text_format is None:
            output_text_format = u"%.18e"
        output_binary = self.has_option([u"-b", u"--binary"])
        output_npz = self.has_option([u"-z", u"--npz"])
        output_npy = self.has_option([u"-n", u"--npy"])
        delete_first = self.has_option([u"-d", u"--delete-first"])
        transpose = self.has_option([u"-t", u"--transpose"])

        self.check_c_extensions("cmfcc")
        if not self.check_input_file(input_file_path):
            return self.ERROR_EXIT_CODE
        if not self.check_output_file(output_file_path):
            return self.ERROR_EXIT_CODE

        try:
            mfccs = AudioFileMFCC(input_file_path, rconf=self.rconf, logger=self.logger).all_mfcc
            if delete_first:
                mfccs = mfccs[1:, :]
            if transpose:
                mfccs = mfccs.transpose()
            if output_binary:
                # save as a raw C float64 binary file
                mapped = numpy.memmap(output_file_path, dtype="float64", mode="w+", shape=mfccs.shape)
                mapped[:] = mfccs[:]
                mapped.flush()
                del mapped
            elif output_npz:
                # save as a .npz compressed binary file
                with io.open(output_file_path, "wb") as output_file:
                    numpy.savez(output_file, mfccs)
            elif output_npy:
                # save as a .npy binary file
                with io.open(output_file_path, "wb") as output_file:
                    numpy.save(output_file, mfccs)
            else:
                # save as a text file
                # NOTE: in Python 2, passing the fmt value a Unicode string crashes NumPy
                #       hence, converting back to bytes, which works in Python 3 too
                numpy.savetxt(output_file_path, mfccs, fmt=gf.safe_bytes(output_text_format))
            self.print_info(u"MFCCs shape: %d %d" % (mfccs.shape))
            self.print_success(u"MFCCs saved to '%s'" % (output_file_path))
            return self.NO_ERROR_EXIT_CODE
        except AudioFileConverterError:
            self.print_error(u"Unable to call the ffmpeg executable '%s'" % (self.rconf[RuntimeConfiguration.FFMPEG_PATH]))
            self.print_error(u"Make sure the path to ffmpeg is correct")
        except (AudioFileUnsupportedFormatError, AudioFileNotInitializedError):
            self.print_error(u"Cannot read file '%s'" % (input_file_path))
            self.print_error(u"Check that its format is supported by ffmpeg")
        except OSError:
            self.print_error(u"Cannot write file '%s'" % (output_file_path))

        return self.ERROR_EXIT_CODE


def main():
    """
    Execute program.
    """
    ExtractMFCCCLI().run(arguments=sys.argv)

if __name__ == '__main__':
    main()



#!/usr/bin/env python
# coding=utf-8

# aeneas is a Python/C library and a set of tools
# to automagically synchronize audio and text (aka forced alignment)
#
# Copyright (C) 2012-2013, Alberto Pettarin (www.albertopettarin.it)
# Copyright (C) 2013-2015, ReadBeyond Srl   (www.readbeyond.it)
# Copyright (C) 2015-2017, Alberto Pettarin (www.albertopettarin.it)
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
This "hydra" tool invokes another aeneas.tool,
according to the specified tool switch.
"""

from __future__ import absolute_import
from __future__ import print_function
import sys

from aeneas.tools.abstract_cli_program import AbstractCLIProgram
from aeneas.tools.convert_syncmap import ConvertSyncMapCLI
from aeneas.tools.download import DownloadCLI
from aeneas.tools.execute_job import ExecuteJobCLI
from aeneas.tools.execute_task import ExecuteTaskCLI
from aeneas.tools.extract_mfcc import ExtractMFCCCLI
from aeneas.tools.ffmpeg_wrapper import FFMPEGWrapperCLI
from aeneas.tools.ffprobe_wrapper import FFPROBEWrapperCLI
from aeneas.tools.plot_waveform import PlotWaveformCLI
from aeneas.tools.read_audio import ReadAudioCLI
from aeneas.tools.read_text import ReadTextCLI
from aeneas.tools.run_sd import RunSDCLI
from aeneas.tools.run_vad import RunVADCLI
from aeneas.tools.synthesize_text import SynthesizeTextCLI
from aeneas.tools.validate import ValidateCLI
import aeneas.globalfunctions as gf


class HydraCLI(AbstractCLIProgram):
    """
    This "hydra" tool invokes another aeneas.tool,
    according to the specified tool switch.
    """
    NAME = gf.file_name_without_extension(__file__)

    HELP = {
        "description": u"Invoke the specified aeneas tool",
        "synopsis": [
            (u"TOOL_PARAMETER TOOL_ARGUMENTS", True)
        ],
        "options": [
        ],
        "parameters": [
            u"--convert-syncmap: call aeneas.tools.convert_syncmap",
            u"--download: call aeneas.tools.download",
            u"--execute-job: call aeneas.tools.execute_job",
            u"--execute-task: call aeneas.tools.execute_task (default)",
            u"--extract-mfcc: call aeneas.tools.extract_mfcc",
            u"--ffmpeg-wrapper: call aeneas.tools.ffmpeg_wrapper",
            u"--ffprobe-wrapper: call aeneas.tools.ffprobe_wrapper",
            u"--plot-waveform: call aeneas.tools.plot_waveform",
            u"--read-audio: call aeneas.tools.read_audio",
            u"--read-text: call aeneas.tools.read_text",
            u"--run-sd: call aeneas.tools.run_sd",
            u"--run-vad: call aeneas.tools.run_vad",
            u"--synthesize-text: call aeneas.tools.synthesize_text",
            u"--validate: call aeneas.tools.validate",
        ],
        "examples": [
            u"--execute-task --help",
            u"--execute-task --examples",
            u"--execute-task --example-json",
            u"--execute-job --help",
        ]
    }

    TOOLS = [
        (ConvertSyncMapCLI, [u"--convert-syncmap"]),
        (DownloadCLI, [u"--download"]),
        (ExecuteJobCLI, [u"--execute-job"]),
        (ExecuteTaskCLI, [u"--execute-task"]),
        (ExtractMFCCCLI, [u"--extract-mfcc"]),
        (FFMPEGWrapperCLI, [u"--ffmpeg-wrapper"]),
        (FFPROBEWrapperCLI, [u"--ffprobe-wrapper"]),
        (PlotWaveformCLI, [u"--plot-waveform"]),
        (ReadAudioCLI, [u"--read-audio"]),
        (ReadTextCLI, [u"--read-text"]),
        (RunSDCLI, [u"--run-sd"]),
        (RunVADCLI, [u"--run-vad"]),
        (SynthesizeTextCLI, [u"--synthesize-text"]),
        (ValidateCLI, [u"--validate"]),
    ]

    def perform_command(self):
        """
        Perform command and return the appropriate exit code.

        :rtype: int
        """
        # if no actual arguments, print help
        if len(self.actual_arguments) < 1:
            return self.print_help(short=True)

        # check if we have a recognized tool switch
        for cls, switches in self.TOOLS:
            if self.has_option(switches):
                arguments = [a for a in sys.argv if a not in switches]
                return cls(invoke=(self.invoke + u" %s" % switches[0])).run(arguments=arguments)

        # check if we have -h, --help, or --version
        if u"-h" in self.actual_arguments:
            return self.print_help(short=True)
        if u"--help" in self.actual_arguments:
            return self.print_help(short=False)
        if u"--version" in self.actual_arguments:
            return self.print_name_version()

        # default to run ExecuteTaskCLI
        return ExecuteTaskCLI(invoke=self.invoke).run(arguments=sys.argv)


def main():
    """
    Execute program.
    """
    HydraCLI().run(arguments=sys.argv, show_help=False)

if __name__ == '__main__':
    main()



#!/usr/bin/env python
# coding=utf-8

# aeneas is a Python/C library and a set of tools
# to automagically synchronize audio and text (aka forced alignment)
#
# Copyright (C) 2012-2013, Alberto Pettarin (www.albertopettarin.it)
# Copyright (C) 2013-2015, ReadBeyond Srl   (www.readbeyond.it)
# Copyright (C) 2015-2017, Alberto Pettarin (www.albertopettarin.it)
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
This module contains the following classes:

* :class:`~aeneas.executejob.ExecuteJob`, a class to process a job;
* :class:`~aeneas.executejob.ExecuteJobExecutionError`,
* :class:`~aeneas.executejob.ExecuteJobInputError`, and
* :class:`~aeneas.executejob.ExecuteJobOutputError`,
  representing errors generated while processing jobs.
"""

from __future__ import absolute_import
from __future__ import print_function

from aeneas.analyzecontainer import AnalyzeContainer
from aeneas.container import Container
from aeneas.container import ContainerFormat
from aeneas.executetask import ExecuteTask
from aeneas.job import Job
from aeneas.logger import Loggable
from aeneas.runtimeconfiguration import RuntimeConfiguration
import aeneas.globalfunctions as gf


class ExecuteJobExecutionError(Exception):
    """
    Error raised when the execution of the job fails for internal reasons.
    """
    pass


class ExecuteJobInputError(Exception):
    """
    Error raised when the input parameters of the job are invalid or missing.
    """
    pass


class ExecuteJobOutputError(Exception):
    """
    Error raised when the creation of the output container failed.
    """
    pass


class ExecuteJob(Loggable):
    """
    Execute a job, that is, execute all of its tasks
    and generate the output container
    holding the generated sync maps.

    If you do not provide a job object in the constructor,
    you must manually set it later, or load it from a container
    with :func:`~aeneas.executejob.ExecuteJob.load_job_from_container`.

    In the first case, you are responsible for setting
    the absolute audio/text/sync map paths of each task of the job,
    to their actual absolute location on the computing machine.
    Moreover, you are responsible for cleaning up
    any temporary files you might have generated around.

    In the second case, you are responsible for
    calling :func:`~aeneas.executejob.ExecuteJob.clean`
    at the end of the job execution,
    to delete the working directory
    created by :func:`~aeneas.executejob.ExecuteJob.load_job_from_container`
    when creating the job object.

    :param job: the job to be executed
    :type  job: :class:`~aeneas.job.Job`
    :param rconf: a runtime configuration
    :type  rconf: :class:`~aeneas.runtimeconfiguration.RuntimeConfiguration`
    :param logger: the logger object
    :type  logger: :class:`~aeneas.logger.Logger`
    :raises: :class:`~aeneas.executejob.ExecuteJobInputError`: if ``job`` is not an instance of ``Job``
    """

    TAG = u"ExecuteJob"

    def __init__(self, job=None, rconf=None, logger=None):
        super(ExecuteJob, self).__init__(rconf=rconf, logger=logger)
        self.job = job
        self.working_directory = None
        self.tmp_directory = None
        if job is not None:
            self.load_job(self.job)

    def load_job(self, job):
        """
        Load the job from the given ``Job`` object.

        :param job: the job to load
        :type  job: :class:`~aeneas.job.Job`
        :raises: :class:`~aeneas.executejob.ExecuteJobInputError`: if ``job`` is not an instance of :class:`~aeneas.job.Job`
        """
        if not isinstance(job, Job):
            self.log_exc(u"job is not an instance of Job", None, True, ExecuteJobInputError)
        self.job = job

    def load_job_from_container(self, container_path, config_string=None):
        """
        Load the job from the given :class:`aeneas.container.Container` object.

        If ``config_string`` is ``None``,
        the container must contain a configuration file;
        otherwise use the provided config string
        (i.e., the wizard case).

        :param string container_path: the path to the input container
        :param string config_string: the configuration string (from wizard)
        :raises: :class:`~aeneas.executejob.ExecuteJobInputError`: if the given container does not contain a valid :class:`~aeneas.job.Job`
        """
        self.log(u"Loading job from container...")

        # create working directory where the input container
        # will be decompressed
        self.working_directory = gf.tmp_directory(root=self.rconf[RuntimeConfiguration.TMP_PATH])
        self.log([u"Created working directory '%s'", self.working_directory])

        try:
            self.log(u"Decompressing input container...")
            input_container = Container(container_path, logger=self.logger)
            input_container.decompress(self.working_directory)
            self.log(u"Decompressing input container... done")
        except Exception as exc:
            self.clean()
            self.log_exc(u"Unable to decompress container '%s': %s" % (container_path, exc), None, True, ExecuteJobInputError)

        try:
            self.log(u"Creating job from working directory...")
            working_container = Container(
                self.working_directory,
                logger=self.logger
            )
            analyzer = AnalyzeContainer(working_container, logger=self.logger)
            self.job = analyzer.analyze(config_string=config_string)
            self.log(u"Creating job from working directory... done")
        except Exception as exc:
            self.clean()
            self.log_exc(u"Unable to analyze container '%s': %s" % (container_path, exc), None, True, ExecuteJobInputError)

        if self.job is None:
            self.log_exc(u"The container '%s' does not contain a valid Job" % (container_path), None, True, ExecuteJobInputError)

        try:
            # set absolute path for text file and audio file
            # for each task in the job
            self.log(u"Setting absolute paths for tasks...")
            for task in self.job.tasks:
                task.text_file_path_absolute = gf.norm_join(
                    self.working_directory,
                    task.text_file_path
                )
                task.audio_file_path_absolute = gf.norm_join(
                    self.working_directory,
                    task.audio_file_path
                )
            self.log(u"Setting absolute paths for tasks... done")

            self.log(u"Loading job from container: succeeded")
        except Exception as exc:
            self.clean()
            self.log_exc(u"Error while setting absolute paths for tasks", exc, True, ExecuteJobInputError)

    def execute(self):
        """
        Execute the job, that is, execute all of its tasks.

        Each produced sync map will be stored
        inside the corresponding task object.

        :raises: :class:`~aeneas.executejob.ExecuteJobExecutionError`: if there is a problem during the job execution
        """
        self.log(u"Executing job")

        if self.job is None:
            self.log_exc(u"The job object is None", None, True, ExecuteJobExecutionError)
        if len(self.job) == 0:
            self.log_exc(u"The job has no tasks", None, True, ExecuteJobExecutionError)
        job_max_tasks = self.rconf[RuntimeConfiguration.JOB_MAX_TASKS]
        if (job_max_tasks > 0) and (len(self.job) > job_max_tasks):
            self.log_exc(u"The Job has %d Tasks, more than the maximum allowed (%d)." % (len(self.job), job_max_tasks), None, True, ExecuteJobExecutionError)
        self.log([u"Number of tasks: '%d'", len(self.job)])

        for task in self.job.tasks:
            try:
                custom_id = task.configuration["custom_id"]
                self.log([u"Executing task '%s'...", custom_id])
                executor = ExecuteTask(task, rconf=self.rconf, logger=self.logger)
                executor.execute()
                self.log([u"Executing task '%s'... done", custom_id])
            except Exception as exc:
                self.log_exc(u"Error while executing task '%s'" % (custom_id), exc, True, ExecuteJobExecutionError)
            self.log(u"Executing task: succeeded")

        self.log(u"Executing job: succeeded")

    def write_output_container(self, output_directory_path):
        """
        Write the output container for this job.

        Return the path to output container,
        which is the concatenation of ``output_directory_path``
        and of the output container file or directory name.

        :param string output_directory_path: the path to a directory where
                                             the output container must be created
        :rtype: string
        :raises: :class:`~aeneas.executejob.ExecuteJobOutputError`: if there is a problem while writing the output container
        """
        self.log(u"Writing output container for this job")

        if self.job is None:
            self.log_exc(u"The job object is None", None, True, ExecuteJobOutputError)
        if len(self.job) == 0:
            self.log_exc(u"The job has no tasks", None, True, ExecuteJobOutputError)
        self.log([u"Number of tasks: '%d'", len(self.job)])

        # create temporary directory where the sync map files
        # will be created
        # this temporary directory will be compressed into
        # the output container
        self.tmp_directory = gf.tmp_directory(root=self.rconf[RuntimeConfiguration.TMP_PATH])
        self.log([u"Created temporary directory '%s'", self.tmp_directory])

        for task in self.job.tasks:
            custom_id = task.configuration["custom_id"]

            # check if the task has sync map and sync map file path
            if task.sync_map_file_path is None:
                self.log_exc(u"Task '%s' has sync_map_file_path not set" % (custom_id), None, True, ExecuteJobOutputError)
            if task.sync_map is None:
                self.log_exc(u"Task '%s' has sync_map not set" % (custom_id), None, True, ExecuteJobOutputError)

            try:
                # output sync map
                self.log([u"Outputting sync map for task '%s'...", custom_id])
                task.output_sync_map_file(self.tmp_directory)
                self.log([u"Outputting sync map for task '%s'... done", custom_id])
            except Exception as exc:
                self.log_exc(u"Error while outputting sync map for task '%s'" % (custom_id), None, True, ExecuteJobOutputError)

        # get output container info
        output_container_format = self.job.configuration["o_container_format"]
        self.log([u"Output container format: '%s'", output_container_format])
        output_file_name = self.job.configuration["o_name"]
        if ((output_container_format != ContainerFormat.UNPACKED) and
                (not output_file_name.endswith(output_container_format))):
            self.log(u"Adding extension to output_file_name")
            output_file_name += "." + output_container_format
        self.log([u"Output file name: '%s'", output_file_name])
        output_file_path = gf.norm_join(
            output_directory_path,
            output_file_name
        )
        self.log([u"Output file path: '%s'", output_file_path])

        try:
            self.log(u"Compressing...")
            container = Container(
                output_file_path,
                output_container_format,
                logger=self.logger
            )
            container.compress(self.tmp_directory)
            self.log(u"Compressing... done")
            self.log([u"Created output file: '%s'", output_file_path])
            self.log(u"Writing output container for this job: succeeded")
            self.clean(False)
            return output_file_path
        except Exception as exc:
            self.clean(False)
            self.log_exc(u"Error while compressing", exc, True, ExecuteJobOutputError)
            return None

    def clean(self, remove_working_directory=True):
        """
        Remove the temporary directory.
        If ``remove_working_directory`` is ``True``
        remove the working directory as well,
        otherwise just remove the temporary directory.

        :param bool remove_working_directory: if ``True``, remove
                                              the working directory as well
        """
        if remove_working_directory is not None:
            self.log(u"Removing working directory... ")
            gf.delete_directory(self.working_directory)
            self.working_directory = None
            self.log(u"Removing working directory... done")
        self.log(u"Removing temporary directory... ")
        gf.delete_directory(self.tmp_directory)
        self.tmp_directory = None
        self.log(u"Removing temporary directory... done")



#!/usr/bin/env python
# coding=utf-8

# aeneas is a Python/C library and a set of tools
# to automagically synchronize audio and text (aka forced alignment)
#
# Copyright (C) 2012-2013, Alberto Pettarin (www.albertopettarin.it)
# Copyright (C) 2013-2015, ReadBeyond Srl   (www.readbeyond.it)
# Copyright (C) 2015-2017, Alberto Pettarin (www.albertopettarin.it)
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
This module contains the following classes:

* :class:`~aeneas.validator.Validator`, assessing whether user input is well-formed;
* :class:`~aeneas.validator.ValidatorResult`, a record holding validation result and possibly messages.
"""

from __future__ import absolute_import
from __future__ import print_function
import io

from aeneas.analyzecontainer import AnalyzeContainer
from aeneas.container import Container
from aeneas.container import ContainerFormat
from aeneas.executetask import AdjustBoundaryAlgorithm
from aeneas.hierarchytype import HierarchyType
from aeneas.idsortingalgorithm import IDSortingAlgorithm
from aeneas.logger import Loggable
from aeneas.runtimeconfiguration import RuntimeConfiguration
from aeneas.syncmap import SyncMapFormat
from aeneas.syncmap import SyncMapHeadTailFormat
from aeneas.textfile import TextFileFormat
import aeneas.globalconstants as gc
import aeneas.globalfunctions as gf


class Validator(Loggable):
    """
    A validator to assess whether user input is well-formed.

    :param rconf: a runtime configuration
    :type  rconf: :class:`~aeneas.runtimeconfiguration.RuntimeConfiguration`
    :param logger: the logger object
    :type  logger: :class:`~aeneas.logger.Logger`
    """

    ALLOWED_VALUES = [
        #
        # NOTE disabling the check on language since now we support multiple TTS
        # COMMENTED (
        # COMMENTED    gc.PPN_JOB_LANGUAGE,
        # COMMENTED    Language.ALLOWED_VALUES
        # COMMENTED ),
        # COMMENTED (
        # COMMENTED     gc.PPN_TASK_LANGUAGE,
        # COMMENTED     Language.ALLOWED_VALUES
        # COMMENTED ),
        #
        (
            gc.PPN_JOB_IS_HIERARCHY_TYPE,
            HierarchyType.ALLOWED_VALUES
        ),
        (
            gc.PPN_JOB_OS_CONTAINER_FORMAT,
            ContainerFormat.ALLOWED_VALUES
        ),
        (
            gc.PPN_JOB_OS_HIERARCHY_TYPE,
            HierarchyType.ALLOWED_VALUES
        ),
        (
            gc.PPN_TASK_IS_TEXT_FILE_FORMAT,
            TextFileFormat.ALLOWED_VALUES
        ),
        (
            gc.PPN_TASK_OS_FILE_FORMAT,
            SyncMapFormat.ALLOWED_VALUES
        ),
        (
            gc.PPN_TASK_IS_TEXT_UNPARSED_ID_SORT,
            IDSortingAlgorithm.ALLOWED_VALUES
        ),
        (
            gc.PPN_TASK_ADJUST_BOUNDARY_ALGORITHM,
            AdjustBoundaryAlgorithm.ALLOWED_VALUES
        ),
        (
            gc.PPN_TASK_OS_FILE_HEAD_TAIL_FORMAT,
            SyncMapHeadTailFormat.ALLOWED_VALUES
        )
    ]

    IMPLIED_PARAMETERS = [
        (
            # is_hierarchy_type=paged => is_task_dir_name_regex
            gc.PPN_JOB_IS_HIERARCHY_TYPE,
            [HierarchyType.PAGED],
            [gc.PPN_JOB_IS_TASK_DIRECTORY_NAME_REGEX]
        ),
        (
            # is_text_type=unparsed => is_text_unparsed_id_sort
            gc.PPN_TASK_IS_TEXT_FILE_FORMAT,
            [TextFileFormat.UNPARSED],
            [gc.PPN_TASK_IS_TEXT_UNPARSED_ID_SORT]
        ),
        (
            # is_text_type=munparsed => is_text_munparsed_l1_id_regex
            gc.PPN_TASK_IS_TEXT_FILE_FORMAT,
            [TextFileFormat.MUNPARSED],
            [gc.PPN_TASK_IS_TEXT_MUNPARSED_L1_ID_REGEX]
        ),
        (
            # is_text_type=munparsed => is_text_munparsed_l2_id_regex
            gc.PPN_TASK_IS_TEXT_FILE_FORMAT,
            [TextFileFormat.MUNPARSED],
            [gc.PPN_TASK_IS_TEXT_MUNPARSED_L2_ID_REGEX]
        ),
        (
            # is_text_type=munparsed => is_text_munparsed_l3_id_regex
            gc.PPN_TASK_IS_TEXT_FILE_FORMAT,
            [TextFileFormat.MUNPARSED],
            [gc.PPN_TASK_IS_TEXT_MUNPARSED_L3_ID_REGEX]
        ),
        (
            # is_text_type=unparsed => is_text_unparsed_class_regex or
            #                          is_text_unparsed_id_regex
            gc.PPN_TASK_IS_TEXT_FILE_FORMAT,
            [TextFileFormat.UNPARSED],
            [
                gc.PPN_TASK_IS_TEXT_UNPARSED_CLASS_REGEX,
                gc.PPN_TASK_IS_TEXT_UNPARSED_ID_REGEX
            ]
        ),
        (
            # os_task_file_format=smil  => os_task_file_smil_audio_ref
            # os_task_file_format=smilh => os_task_file_smil_audio_ref
            # os_task_file_format=smilm => os_task_file_smil_audio_ref
            gc.PPN_TASK_OS_FILE_FORMAT,
            [SyncMapFormat.SMIL, SyncMapFormat.SMILH, SyncMapFormat.SMILM],
            [gc.PPN_TASK_OS_FILE_SMIL_AUDIO_REF]
        ),
        (
            # os_task_file_format=smil  => os_task_file_smil_page_ref
            # os_task_file_format=smilh => os_task_file_smil_page_ref
            # os_task_file_format=smilm => os_task_file_smil_page_ref
            gc.PPN_TASK_OS_FILE_FORMAT,
            [SyncMapFormat.SMIL, SyncMapFormat.SMILH, SyncMapFormat.SMILM],
            [gc.PPN_TASK_OS_FILE_SMIL_PAGE_REF]
        ),
        (
            # task_adjust_boundary_algorithm=percent => task_adjust_boundary_percent_value
            gc.PPN_TASK_ADJUST_BOUNDARY_ALGORITHM,
            [AdjustBoundaryAlgorithm.PERCENT],
            [gc.PPN_TASK_ADJUST_BOUNDARY_PERCENT_VALUE]
        ),
        (
            # task_adjust_boundary_algorithm=rate => task_adjust_boundary_rate_value
            gc.PPN_TASK_ADJUST_BOUNDARY_ALGORITHM,
            [AdjustBoundaryAlgorithm.RATE],
            [gc.PPN_TASK_ADJUST_BOUNDARY_RATE_VALUE]
        ),
        (
            # task_adjust_boundary_algorithm=rate_aggressive => task_adjust_boundary_rate_value
            gc.PPN_TASK_ADJUST_BOUNDARY_ALGORITHM,
            [AdjustBoundaryAlgorithm.RATEAGGRESSIVE],
            [gc.PPN_TASK_ADJUST_BOUNDARY_RATE_VALUE]
        ),
        (
            # task_adjust_boundary_algorithm=currentend => task_adjust_boundary_currentend_value
            gc.PPN_TASK_ADJUST_BOUNDARY_ALGORITHM,
            [AdjustBoundaryAlgorithm.AFTERCURRENT],
            [gc.PPN_TASK_ADJUST_BOUNDARY_AFTERCURRENT_VALUE]
        ),
        (
            # task_adjust_boundary_algorithm=rate => task_adjust_boundary_nextstart_value
            gc.PPN_TASK_ADJUST_BOUNDARY_ALGORITHM,
            [AdjustBoundaryAlgorithm.BEFORENEXT],
            [gc.PPN_TASK_ADJUST_BOUNDARY_BEFORENEXT_VALUE]
        ),
        (
            # task_adjust_boundary_algorithm=offset => task_adjust_boundary_offset_value
            gc.PPN_TASK_ADJUST_BOUNDARY_ALGORITHM,
            [AdjustBoundaryAlgorithm.OFFSET],
            [gc.PPN_TASK_ADJUST_BOUNDARY_OFFSET_VALUE]
        )
    ]

    JOB_REQUIRED_PARAMETERS = [
        gc.PPN_JOB_LANGUAGE,
        gc.PPN_JOB_OS_CONTAINER_FORMAT,
        gc.PPN_JOB_OS_FILE_NAME,
    ]

    TASK_REQUIRED_PARAMETERS = [
        gc.PPN_TASK_IS_TEXT_FILE_FORMAT,
        gc.PPN_TASK_LANGUAGE,
        gc.PPN_TASK_OS_FILE_FORMAT,
        gc.PPN_TASK_OS_FILE_NAME,
    ]

    TASK_REQUIRED_PARAMETERS_EXTERNAL_NAME = [
        gc.PPN_TASK_IS_TEXT_FILE_FORMAT,
        gc.PPN_TASK_LANGUAGE,
        gc.PPN_TASK_OS_FILE_FORMAT,
    ]

    TXT_REQUIRED_PARAMETERS = [
        gc.PPN_JOB_IS_AUDIO_FILE_NAME_REGEX,
        gc.PPN_JOB_IS_AUDIO_FILE_RELATIVE_PATH,
        gc.PPN_JOB_IS_HIERARCHY_PREFIX,
        gc.PPN_JOB_IS_HIERARCHY_TYPE,
        gc.PPN_JOB_IS_TEXT_FILE_NAME_REGEX,
        gc.PPN_JOB_IS_TEXT_FILE_RELATIVE_PATH,
        gc.PPN_JOB_LANGUAGE,
        gc.PPN_JOB_OS_CONTAINER_FORMAT,
        gc.PPN_JOB_OS_FILE_NAME,
        gc.PPN_JOB_OS_HIERARCHY_PREFIX,
        gc.PPN_JOB_OS_HIERARCHY_TYPE,
        gc.PPN_TASK_IS_TEXT_FILE_FORMAT,
        gc.PPN_TASK_OS_FILE_FORMAT,
        gc.PPN_TASK_OS_FILE_NAME,
    ]

    XML_JOB_REQUIRED_PARAMETERS = [
        gc.PPN_JOB_OS_CONTAINER_FORMAT,
        gc.PPN_JOB_OS_FILE_NAME,
        gc.PPN_JOB_OS_HIERARCHY_PREFIX,
        gc.PPN_JOB_OS_HIERARCHY_TYPE,
    ]

    XML_TASK_REQUIRED_PARAMETERS = [
        gc.PPN_TASK_IS_AUDIO_FILE_XML,
        gc.PPN_TASK_IS_TEXT_FILE_FORMAT,
        gc.PPN_TASK_IS_TEXT_FILE_XML,
        gc.PPN_TASK_LANGUAGE,
        gc.PPN_TASK_OS_FILE_FORMAT,
        gc.PPN_TASK_OS_FILE_NAME,
    ]

    TAG = u"Validator"

    def __init__(self, rconf=None, logger=None):
        super(Validator, self).__init__(rconf=rconf, logger=logger)
        self.result = None

    def check_file_encoding(self, input_file_path):
        """
        Check whether the given file is UTF-8 encoded.

        :param string input_file_path: the path of the file to be checked
        :rtype: :class:`~aeneas.validator.ValidatorResult`
        """
        self.log([u"Checking encoding of file '%s'", input_file_path])
        self.result = ValidatorResult()
        if self._are_safety_checks_disabled(u"check_file_encoding"):
            return self.result
        if not gf.file_can_be_read(input_file_path):
            self._failed(u"File '%s' cannot be read." % (input_file_path))
            return self.result
        with io.open(input_file_path, "rb") as file_object:
            bstring = file_object.read()
            self._check_utf8_encoding(bstring)
        return self.result

    def check_raw_string(self, string, is_bstring=True):
        """
        Check whether the given string
        is properly UTF-8 encoded (if ``is_bytes`` is ``True``),
        it is not empty, and
        it does not contain reserved characters.

        :param string string: the byte string or Unicode string to be checked
        :param bool is_bstring: if True, string is a byte string
        :rtype: :class:`~aeneas.validator.ValidatorResult`
        """
        self.log(u"Checking the given byte string")
        self.result = ValidatorResult()
        if self._are_safety_checks_disabled(u"check_raw_string"):
            return self.result
        if is_bstring:
            self._check_utf8_encoding(string)
            if not self.result.passed:
                return self.result
            string = gf.safe_unicode(string)
        self._check_not_empty(string)
        if not self.result.passed:
            return self.result
        self._check_reserved_characters(string)
        return self.result

    def check_configuration_string(
            self,
            config_string,
            is_job=True,
            external_name=False
    ):
        """
        Check whether the given job or task configuration string
        is well-formed (if ``is_bstring`` is ``True``)
        and it has all the required parameters.

        :param string config_string: the byte string or Unicode string to be checked
        :param bool is_job: if ``True``, ``config_string`` is a job config string
        :param bool external_name: if ``True``, the task name is provided externally,
                                   and it is not required to appear
                                   in the config string
        :rtype: :class:`~aeneas.validator.ValidatorResult`
        """
        if is_job:
            self.log(u"Checking job configuration string")
        else:
            self.log(u"Checking task configuration string")
        self.result = ValidatorResult()
        if self._are_safety_checks_disabled(u"check_configuration_string"):
            return self.result
        if is_job:
            required_parameters = self.JOB_REQUIRED_PARAMETERS
        elif external_name:
            required_parameters = self.TASK_REQUIRED_PARAMETERS_EXTERNAL_NAME
        else:
            required_parameters = self.TASK_REQUIRED_PARAMETERS
        is_bstring = gf.is_bytes(config_string)
        if is_bstring:
            self.log(u"Checking that config_string is well formed")
            self.check_raw_string(config_string, is_bstring=True)
            if not self.result.passed:
                return self.result
            config_string = gf.safe_unicode(config_string)
        self.log(u"Checking required parameters")
        parameters = gf.config_string_to_dict(config_string, self.result)
        self._check_required_parameters(required_parameters, parameters)
        self.log([u"Checking config_string: returning %s", self.result.passed])
        return self.result

    def check_config_txt(self, contents, is_config_string=False):
        """
        Check whether the given TXT config file contents
        (if ``is_config_string`` is ``False``) or
        TXT config string (if ``is_config_string`` is ``True``)
        is well-formed and it has all the required parameters.

        :param string contents: the TXT config file contents or TXT config string
        :param bool is_config_string: if ``True``, contents is a config string
        :rtype: :class:`~aeneas.validator.ValidatorResult`
        """
        self.log(u"Checking contents TXT config file")
        self.result = ValidatorResult()
        if self._are_safety_checks_disabled(u"check_config_txt"):
            return self.result
        is_bstring = gf.is_bytes(contents)
        if is_bstring:
            self.log(u"Checking that contents is well formed")
            self.check_raw_string(contents, is_bstring=True)
            if not self.result.passed:
                return self.result
            contents = gf.safe_unicode(contents)
        if not is_config_string:
            self.log(u"Converting file contents to config string")
            contents = gf.config_txt_to_string(contents)
        self.log(u"Checking required parameters")
        required_parameters = self.TXT_REQUIRED_PARAMETERS
        parameters = gf.config_string_to_dict(contents, self.result)
        self._check_required_parameters(required_parameters, parameters)
        self.log([u"Checking contents: returning %s", self.result.passed])
        return self.result

    def check_config_xml(self, contents):
        """
        Check whether the given XML config file contents
        is well-formed and it has all the required parameters.

        :param string contents: the XML config file contents or XML config string
        :param bool is_config_string: if ``True``, contents is a config string
        :rtype: :class:`~aeneas.validator.ValidatorResult`
        """
        self.log(u"Checking contents XML config file")
        self.result = ValidatorResult()
        if self._are_safety_checks_disabled(u"check_config_xml"):
            return self.result
        contents = gf.safe_bytes(contents)
        self.log(u"Checking that contents is well formed")
        self.check_raw_string(contents, is_bstring=True)
        if not self.result.passed:
            return self.result
        self.log(u"Checking required parameters for job")
        job_parameters = gf.config_xml_to_dict(contents, self.result, parse_job=True)
        self._check_required_parameters(self.XML_JOB_REQUIRED_PARAMETERS, job_parameters)
        if not self.result.passed:
            return self.result
        self.log(u"Checking required parameters for task")
        tasks_parameters = gf.config_xml_to_dict(contents, self.result, parse_job=False)
        for parameters in tasks_parameters:
            self.log([u"Checking required parameters for task: '%s'", parameters])
            self._check_required_parameters(self.XML_TASK_REQUIRED_PARAMETERS, parameters)
            if not self.result.passed:
                return self.result
        return self.result

    def check_container(self, container_path, container_format=None, config_string=None):
        """
        Check whether the given container is well-formed.

        :param string container_path: the path of the container to be checked
        :param container_format: the format of the container
        :type  container_format: :class:`~aeneas.container.ContainerFormat`
        :param string config_string: the configuration string generated by the wizard
        :rtype: :class:`~aeneas.validator.ValidatorResult`
        """
        self.log([u"Checking container '%s'", container_path])
        self.result = ValidatorResult()

        if self._are_safety_checks_disabled(u"check_container"):
            return self.result

        if not (gf.file_exists(container_path) or gf.directory_exists(container_path)):
            self._failed(u"Container '%s' not found." % container_path)
            return self.result

        container = Container(container_path, container_format)
        try:
            self.log(u"Checking container has config file")
            if config_string is not None:
                self.log(u"Container with config string from wizard")
                self.check_config_txt(config_string, is_config_string=True)
            elif container.has_config_xml:
                self.log(u"Container has XML config file")
                contents = container.read_entry(container.entry_config_xml)
                if contents is None:
                    self._failed(u"Unable to read the contents of XML config file.")
                    return self.result
                self.check_config_xml(contents)
            elif container.has_config_txt:
                self.log(u"Container has TXT config file")
                contents = container.read_entry(container.entry_config_txt)
                if contents is None:
                    self._failed(u"Unable to read the contents of TXT config file.")
                    return self.result
                self.check_config_txt(contents, is_config_string=False)
            else:
                self._failed(u"Container does not have a TXT or XML configuration file.")

            self.log(u"Checking we have a valid job in the container")
            if not self.result.passed:
                return self.result
            self.log(u"Analyze the contents of the container")
            analyzer = AnalyzeContainer(container)
            if config_string is not None:
                job = analyzer.analyze(config_string=config_string)
            else:
                job = analyzer.analyze()
            self._check_analyzed_job(job, container)

        except OSError:
            self._failed(u"Unable to read the contents of the container.")
        return self.result

    def _are_safety_checks_disabled(self, caller=u"unknown_function"):
        """
        Return ``True`` if safety checks are disabled.

        :param string caller: the name of the caller function
        :rtype: bool
        """
        if self.rconf.safety_checks:
            return False
        self.log_warn([u"Safety checks disabled => %s passed", caller])
        return True

    def _failed(self, msg):
        """
        Log a validation failure.

        :param string msg: the error message
        """
        self.log(msg)
        self.result.passed = False
        self.result.add_error(msg)
        self.log(u"Failed")

    def _check_utf8_encoding(self, bstring):
        """
        Check whether the given sequence of bytes
        is properly encoded in UTF-8.

        :param bytes bstring: the byte string to be checked
        """
        if not gf.is_bytes(bstring):
            self._failed(u"The given string is not a sequence of bytes")
            return
        if not gf.is_utf8_encoded(bstring):
            self._failed(u"The given string is not encoded in UTF-8.")

    def _check_not_empty(self, string):
        """
        Check whether the given string has zero length.

        :param string string: the byte string or Unicode string to be checked
        """
        if len(string) == 0:
            self._failed(u"The given string has zero length")

    def _check_reserved_characters(self, ustring):
        """
        Check whether the given Unicode string contains reserved characters.

        :param string ustring: the string to be checked
        """
        forbidden = [c for c in gc.CONFIG_RESERVED_CHARACTERS if c in ustring]
        if len(forbidden) > 0:
            self._failed(u"The given string contains the reserved characters '%s'." % u" ".join(forbidden))

    def _check_allowed_values(self, parameters):
        """
        Check whether the given parameter value is allowed.
        Log messages into ``self.result``.

        :param dict parameters: the given parameters
        """
        for key, allowed_values in self.ALLOWED_VALUES:
            self.log([u"Checking allowed values for parameter '%s'", key])
            if key in parameters:
                value = parameters[key]
                if value not in allowed_values:
                    self._failed(u"Parameter '%s' has value '%s' which is not allowed." % (key, value))
                    return
        self.log(u"Passed")

    def _check_implied_parameters(self, parameters):
        """
        Check whether at least one of the keys in implied_keys
        is in ``parameters``,
        when a given ``key=value`` is present in ``parameters``,
        for some value in values.
        Log messages into ``self.result``.

        :param dict parameters: the given parameters
        """
        for key, values, implied_keys in self.IMPLIED_PARAMETERS:
            self.log([u"Checking implied parameters by '%s'='%s'", key, values])
            if (key in parameters) and (parameters[key] in values):
                found = False
                for implied_key in implied_keys:
                    if implied_key in parameters:
                        found = True
                if not found:
                    if len(implied_keys) == 1:
                        msg = u"Parameter '%s' is required when '%s'='%s'." % (implied_keys[0], key, parameters[key])
                    else:
                        msg = u"At least one of [%s] is required when '%s'='%s'." % (",".join(implied_keys), key, parameters[key])
                    self._failed(msg)
                    return
        self.log(u"Passed")

    def _check_required_parameters(
            self,
            required_parameters,
            parameters
    ):
        """
        Check whether the given parameter dictionary contains
        all the required paramenters.
        Log messages into ``self.result``.

        :param list required_parameters: required parameters
        :param dict parameters: parameters specified by the user
        """
        self.log([u"Checking required parameters '%s'", required_parameters])
        self.log(u"Checking input parameters are not empty")
        if (parameters is None) or (len(parameters) == 0):
            self._failed(u"No parameters supplied.")
            return
        self.log(u"Checking no required parameter is missing")
        for req_param in required_parameters:
            if req_param not in parameters:
                self._failed(u"Required parameter '%s' not set." % req_param)
                return
        self.log(u"Checking all parameter values are allowed")
        self._check_allowed_values(parameters)
        self.log(u"Checking all implied parameters are present")
        self._check_implied_parameters(parameters)
        return self.result

    def _check_analyzed_job(self, job, container):
        """
        Check that the job object generated from the given container
        is well formed, that it has at least one task,
        and that the text file of each task has the correct encoding.
        Log messages into ``self.result``.

        :param job: the Job object generated from container
        :type  job: :class:`~aeneas.job.Job`
        :param container: the Container object
        :type  container: :class:`~aeneas.container.Container`
        """
        self.log(u"Checking the Job object generated from container")

        self.log(u"Checking that the Job is not None")
        if job is None:
            self._failed(u"Unable to create a Job from the container.")
            return

        self.log(u"Checking that the Job has at least one Task")
        if len(job) == 0:
            self._failed(u"Unable to create at least one Task from the container.")
            return

        if self.rconf[RuntimeConfiguration.JOB_MAX_TASKS] > 0:
            self.log(u"Checking that the Job does not have too many Tasks")
            if len(job) > self.rconf[RuntimeConfiguration.JOB_MAX_TASKS]:
                self._failed(u"The Job has %d Tasks, more than the maximum allowed (%d)." % (
                    len(job),
                    self.rconf[RuntimeConfiguration.JOB_MAX_TASKS]
                ))
                return

        self.log(u"Checking that each Task text file is well formed")
        for task in job.tasks:
            self.log([u"Checking Task text file '%s'", task.text_file_path])
            text_file_bstring = container.read_entry(task.text_file_path)
            if (text_file_bstring is None) or (len(text_file_bstring) == 0):
                self._failed(u"Text file '%s' is empty" % task.text_file_path)
                return
            self._check_utf8_encoding(text_file_bstring)
            if not self.result.passed:
                self._failed(u"Text file '%s' is not encoded in UTF-8" % task.text_file_path)
                return
            self._check_not_empty(text_file_bstring)
            if not self.result.passed:
                self._failed(u"Text file '%s' is empty" % task.text_file_path)
                return
            self.log([u"Checking Task text file '%s': passed", task.text_file_path])
        self.log(u"Checking each Task text file is well formed: passed")


class ValidatorResult(object):
    """
    A structure to contain the result of a validation.
    """

    TAG = u"ValidatorResult"

    def __init__(self):
        self.passed = True
        self.warnings = []
        self.errors = []

    def __unicode__(self):
        msg = [
            u"Passed: %s" % self.passed,
            self.pretty_print(warnings=True)
        ]
        return u"\n".join(msg)

    def __str__(self):
        return gf.safe_str(self.__unicode__())

    def pretty_print(self, warnings=False):
        """
        Pretty print warnings and errors.

        :param bool warnings: if ``True``, also print warnings.
        :rtype: string
        """
        msg = []
        if (warnings) and (len(self.warnings) > 0):
            msg.append(u"Warnings:")
            for warning in self.warnings:
                msg.append(u"  %s" % warning)
        if len(self.errors) > 0:
            msg.append(u"Errors:")
            for error in self.errors:
                msg.append(u"  %s" % error)
        return u"\n".join(msg)

    @property
    def passed(self):
        """
        The result of a validation.

        Return ``True`` if passed, possibly with emitted warnings.

        Return ``False`` if not passed, that is, at least one error emitted.

        :rtype: bool
        """
        return self.__passed

    @passed.setter
    def passed(self, passed):
        self.__passed = passed

    @property
    def warnings(self):
        """
        The list of emitted warnings.

        :rtype: list of strings
        """
        return self.__warnings

    @warnings.setter
    def warnings(self, warnings):
        self.__warnings = warnings

    @property
    def errors(self):
        """
        The list of emitted errors.

        :rtype: list of strings
        """
        return self.__errors

    @errors.setter
    def errors(self, errors):
        self.__errors = errors

    def add_warning(self, message):
        """
        Add a message to the warnings.

        :param string message: the message to be added
        """
        self.warnings.append(message)

    def add_error(self, message):
        """
        Add a message to the errors.

        :param string message: the message to be added
        """
        self.errors.append(message)



#!/usr/bin/env python
# coding=utf-8

# aeneas is a Python/C library and a set of tools
# to automagically synchronize audio and text (aka forced alignment)
#
# Copyright (C) 2012-2013, Alberto Pettarin (www.albertopettarin.it)
# Copyright (C) 2013-2015, ReadBeyond Srl   (www.readbeyond.it)
# Copyright (C) 2015-2017, Alberto Pettarin (www.albertopettarin.it)
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
This module contains the following classes:

* :class:`~aeneas.adjustboundaryalgorithm.AdjustBoundaryAlgorithm`
  implementing functions to adjust
  the boundary point between two consecutive fragments.

.. warning:: This module is likely to be refactored in a future version
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from aeneas.audiofilemfcc import AudioFileMFCC
from aeneas.exacttiming import Decimal
from aeneas.exacttiming import TimeInterval
from aeneas.exacttiming import TimeValue
from aeneas.logger import Loggable
from aeneas.runtimeconfiguration import RuntimeConfiguration
from aeneas.syncmap import SyncMapFragment
from aeneas.syncmap import SyncMapFragmentList
from aeneas.textfile import TextFile
from aeneas.textfile import TextFragment
from aeneas.tree import Tree
import aeneas.globalconstants as gc


class AdjustBoundaryAlgorithm(Loggable):
    """
    Enumeration and implementation of the available algorithms
    to adjust the boundary point between two consecutive fragments.

    :param rconf: a runtime configuration
    :type  rconf: :class:`~aeneas.runtimeconfiguration.RuntimeConfiguration`
    :param logger: the logger object
    :type  logger: :class:`~aeneas.logger.Logger`
    :raises: TypeError: if one of ``boundary_indices``, ``real_wave_mfcc``,
                        or ``text_file`` is ``None`` or it has a wrong type
    """

    AFTERCURRENT = "aftercurrent"
    """
    Set the boundary at ``value`` seconds
    after the end of the current fragment,
    if the current boundary falls inside
    a nonspeech interval.
    If not, no adjustment is made.

    Example (value ``0.200`` seconds):

    .. image:: _static/aftercurrent.200.png
       :scale: 100%
       :align: center
       :alt: Comparison between AUTO labels and AFTERCURRENT labels with 0.200 seconds offset
    """

    AUTO = "auto"
    """
    Auto (no adjustment).

    Example:

    .. image:: _static/auto.png
       :scale: 100%
       :align: center
       :alt: The AUTO method does not change the time intervals
    """

    BEFORENEXT = "beforenext"
    """
    Set the boundary at ``value`` seconds
    before the beginning of the next fragment,
    if the current boundary falls inside
    a nonspeech interval.
    If not, no adjustment is made.

    Example (value ``0.200`` seconds):

    .. image:: _static/beforenext.200.png
       :scale: 100%
       :align: center
       :alt: Comparison between AUTO labels and BEFORENEXT labels with 0.200 seconds offset
    """

    OFFSET = "offset"
    """
    Offset the current boundaries by ``value`` seconds.
    The ``value`` can be negative or positive.

    Example (value ``-0.200`` seconds):

    .. image:: _static/offset.m200.png
       :scale: 100%
       :align: center
       :alt: Comparison between AUTO labels and OFFSET labels with value -0.200

    Example (value ``0.200`` seconds):

    .. image:: _static/offset.200.png
       :scale: 100%
       :align: center
       :alt: Comparison between AUTO labels and OFFSET labels with value 0.200

    .. versionadded:: 1.1.0
    """

    PERCENT = "percent"
    """
    Set the boundary at ``value`` percent of
    the nonspeech interval between the current and the next fragment,
    if the current boundary falls inside
    a nonspeech interval.
    The ``value`` must be an integer in ``[0, 100]``.
    If not, no adjustment is made.

    Example (value ``25`` %):

    .. image:: _static/percent.25.png
       :scale: 100%
       :align: center
       :alt: Comparison between AUTO labels and PERCENT labels with value 25 %

    Example (value ``50`` %):

    .. image:: _static/percent.50.png
       :scale: 100%
       :align: center
       :alt: Comparison between AUTO labels and PERCENT labels with value 50 %

    Example (value ``75`` %):

    .. image:: _static/percent.75.png
       :scale: 100%
       :align: center
       :alt: Comparison between AUTO labels and PERCENT labels with value 75 %

    """

    RATE = "rate"
    """
    Adjust boundaries trying to respect the
    ``value`` characters/second constraint.
    The ``value`` must be positive.
    First, the rates of all fragments are computed,
    using the current boundaries.
    For those fragments exceeding ``value`` characters/second,
    the algorithm will try to move the end boundary forward,
    so that its time interval increases (and hence its rate decreases).
    Clearly, it is possible that not all fragments
    can be adjusted this way: for example,
    if you have three consecutive fragments exceeding ``value``,
    the middle one cannot be stretched.

    Example (value ``13.0``, note how ``f000003`` is modified):

    .. image:: _static/rate.13.png
       :scale: 100%
       :align: center
       :alt: Comparison between AUTO labels and RATE labels with value 13.0

    """

    RATEAGGRESSIVE = "rateaggressive"
    """
    Adjust boundaries trying to respect the
    ``value`` characters/second constraint, in aggressive mode.
    The ``value`` must be positive.
    First, the rates of all fragments are computed,
    using the current boundaries.
    For those fragments exceeding ``value`` characters/second,
    the algorithm will try to move the end boundary forward,
    so that its time interval increases (and hence its rate decreases).
    If moving the end boundary is not possible,
    or it is not enough to keep the rate below ``value``,
    the algorithm will try to move the begin boundary back;
    this is the difference with the less aggressive
    :data:`~aeneas.adjustboundaryalgorithm.AdjustBoundaryAlgorithm.RATE`
    algorithm.
    Clearly, it is possible that not all fragments
    can be adjusted this way: for example,
    if you have three consecutive fragments exceeding ``value``,
    the middle one cannot be stretched.

    Example (value ``13.0``, note how ``f000003`` is modified):

    .. image:: _static/rateaggressive.13.png
       :scale: 100%
       :align: center
       :alt: Comparison between AUTO labels and RATEAGGRESSIVE labels with value 13.0

    .. versionadded:: 1.1.0
    """

    ALLOWED_VALUES = [
        AFTERCURRENT,
        AUTO,
        BEFORENEXT,
        OFFSET,
        PERCENT,
        RATE,
        RATEAGGRESSIVE
    ]
    """ List of all the allowed values """

    TAG = u"AdjustBoundaryAlgorithm"

    def __init__(self, rconf=None, logger=None):
        super(AdjustBoundaryAlgorithm, self).__init__(rconf=rconf, logger=logger)
        self.smflist = None
        self.mws = self.rconf.mws

    def adjust(
        self,
        aba_parameters,
        boundary_indices,
        real_wave_mfcc,
        text_file,
        allow_arbitrary_shift=False
    ):
        """
        Adjust the boundaries of the text map
        using the algorithm and parameters
        specified in the constructor,
        storing the sync map fragment list internally.

        :param dict aba_parameters: a dictionary containing the algorithm and its parameters,
                                    as produced by ``aba_parameters()`` in ``TaskConfiguration``
        :param boundary_indices: the current boundary indices,
                                 with respect to the audio file full MFCCs
        :type  boundary_indices: :class:`numpy.ndarray` (1D)
        :param real_wave_mfcc: the audio file MFCCs
        :type  real_wave_mfcc: :class:`~aeneas.audiofilemfcc.AudioFileMFCC`
        :param text_file: the text file containing the text fragments associated
        :type  text_file: :class:`~aeneas.textfile.TextFile`
        :param bool allow_arbitrary_shift: if ``True``, allow arbitrary shifts when adjusting zero length

        :rtype: list of :class:`~aeneas.syncmap.SyncMapFragmentList`
        """
        self.log(u"Called adjust")
        if boundary_indices is None:
            self.log_exc(u"boundary_indices is None", None, True, TypeError)
        if not isinstance(real_wave_mfcc, AudioFileMFCC):
            self.log_exc(u"real_wave_mfcc is not an AudioFileMFCC object", None, True, TypeError)
        if not isinstance(text_file, TextFile):
            self.log_exc(u"text_file is not a TextFile object", None, True, TypeError)

        nozero = aba_parameters["nozero"]
        ns_min, ns_string = aba_parameters["nonspeech"]
        algorithm, algo_parameters = aba_parameters["algorithm"]

        self.log(u"  Converting boundary indices to fragment list...")
        begin = real_wave_mfcc.middle_begin * real_wave_mfcc.rconf.mws
        end = real_wave_mfcc.middle_end * real_wave_mfcc.rconf.mws
        time_values = [begin] + list(boundary_indices * self.mws) + [end]
        self.intervals_to_fragment_list(
            text_file=text_file,
            time_values=time_values
        )
        self.log(u"  Converting boundary indices to fragment list... done")

        self.log(u"  Processing fragments with zero length...")
        self._process_zero_length(nozero, allow_arbitrary_shift)
        self.log(u"  Processing fragments with zero length... done")

        self.log(u"  Processing nonspeech fragments...")
        self._process_long_nonspeech(ns_min, ns_string, real_wave_mfcc)
        self.log(u"  Processing nonspeech fragments... done")

        self.log(u"  Adjusting...")
        ALGORITHM_MAP = {
            self.AFTERCURRENT: self._adjust_aftercurrent,
            self.AUTO: self._adjust_auto,
            self.BEFORENEXT: self._adjust_beforenext,
            self.OFFSET: self._adjust_offset,
            self.PERCENT: self._adjust_percent,
            self.RATE: self._adjust_rate,
            self.RATEAGGRESSIVE: self._adjust_rate_aggressive,
        }
        ALGORITHM_MAP[algorithm](real_wave_mfcc, algo_parameters)
        self.log(u"  Adjusting... done")

        self.log(u"  Smoothing...")
        self._smooth_fragment_list(real_wave_mfcc.audio_length, ns_string)
        self.log(u"  Smoothing... done")

        return self.smflist

    def intervals_to_fragment_list(self, text_file, time_values):
        """
        Transform a list of at least 4 time values
        (corresponding to at least 3 intervals)
        into a sync map fragment list and store it internally.
        The first interval is a HEAD, the last is a TAIL.

        For example:

            time_values=[0.000, 1.000, 2.000, 3.456] => [(0.000, 1.000), (1.000, 2.000), (2.000, 3.456)]

        :param text_file: the text file containing the text fragments associated
        :type  text_file: :class:`~aeneas.textfile.TextFile`
        :param time_values: the time values
        :type  time_values: list of :class:`~aeneas.exacttiming.TimeValue`
        :raises: TypeError: if ``text_file`` is not an instance of :class:`~aeneas.textfile.TextFile`
                            or ``time_values`` is not a list
        :raises: ValueError: if ``time_values`` has length less than four
        """
        if not isinstance(text_file, TextFile):
            self.log_exc(u"text_file is not an instance of TextFile", None, True, TypeError)
        if not isinstance(time_values, list):
            self.log_exc(u"time_values is not a list", None, True, TypeError)
        if len(time_values) < 4:
            self.log_exc(u"time_values has length < 4", None, True, ValueError)
        self.log(u"Converting time values to fragment list...")
        begin = time_values[0]
        end = time_values[-1]
        self.log([u"  Creating SyncMapFragmentList with begin %.3f and end %.3f", begin, end])
        self.smflist = SyncMapFragmentList(
            begin=begin,
            end=end,
            rconf=self.rconf,
            logger=self.logger
        )
        self.log(u"  Creating HEAD fragment")
        self.smflist.add(SyncMapFragment(
            # NOTE lines and filtered lines MUST be set,
            #      otherwise some output format might break
            #      when adding HEAD/TAIL to output
            text_fragment=TextFragment(identifier=u"HEAD", lines=[], filtered_lines=[]),
            begin=time_values[0],
            end=time_values[1],
            fragment_type=SyncMapFragment.HEAD
        ), sort=False)
        self.log(u"  Creating REGULAR fragments")
        # NOTE text_file.fragments() returns a list,
        #      so we cache a copy here instead of
        #      calling it once per loop
        fragments = text_file.fragments
        for i in range(1, len(time_values) - 2):
            self.log([u"    Adding fragment %d ...", i])
            self.smflist.add(SyncMapFragment(
                text_fragment=fragments[i - 1],
                begin=time_values[i],
                end=time_values[i + 1],
                fragment_type=SyncMapFragment.REGULAR
            ), sort=False)
            self.log([u"    Adding fragment %d ... done", i])
        self.log(u"  Creating TAIL fragment")
        self.smflist.add(SyncMapFragment(
            # NOTE lines and filtered lines MUST be set,
            #      otherwise some output format might break
            #      when adding HEAD/TAIL to output
            text_fragment=TextFragment(identifier=u"TAIL", lines=[], filtered_lines=[]),
            begin=time_values[len(time_values) - 2],
            end=end,
            fragment_type=SyncMapFragment.TAIL
        ), sort=False)
        self.log(u"Converting time values to fragment list... done")
        self.log(u"Sorting fragment list...")
        self.smflist.sort()
        self.log(u"Sorting fragment list... done")
        return self.smflist

    def append_fragment_list_to_sync_root(self, sync_root):
        """
        Append the sync map fragment list
        to the given node from a sync map tree.

        :param sync_root: the root of the sync map tree to which the new nodes should be appended
        :type  sync_root: :class:`~aeneas.tree.Tree`
        """
        if not isinstance(sync_root, Tree):
            self.log_exc(u"sync_root is not a Tree object", None, True, TypeError)

        self.log(u"Appending fragment list to sync root...")
        for fragment in self.smflist:
            sync_root.add_child(Tree(value=fragment))
        self.log(u"Appending fragment list to sync root... done")

    # #####################################################
    # NO ZERO AND LONG NONSPEECH FUNCTIONS
    # #####################################################

    def _process_zero_length(self, nozero, allow_arbitrary_shift):
        """
        If ``nozero`` is ``True``, modify the sync map fragment list
        so that no fragment will have zero length.
        """
        self.log(u"Called _process_zero_length")
        if not nozero:
            self.log(u"Processing zero length intervals not requested: returning")
            return
        self.log(u"Processing zero length intervals requested")
        self.log(u"  Checking and fixing...")
        duration = self.rconf[RuntimeConfiguration.ABA_NO_ZERO_DURATION]
        self.log([u"  Requested no zero duration: %.3f", duration])
        if not allow_arbitrary_shift:
            self.log(u"  No arbitrary shift => taking max with mws")
            duration = self.rconf.mws.geq_multiple(duration)
        self.log([u"  Actual no zero duration: %.3f", duration])
        # ignore HEAD and TAIL
        max_index = len(self.smflist) - 1
        self.smflist.fix_zero_length_fragments(
            duration=duration,
            min_index=1,
            max_index=max_index
        )
        self.log(u"  Checking and fixing... done")
        if self.smflist.has_zero_length_fragments(1, max_index):
            self.log_warn(u"  The fragment list still has fragments with zero length")
        else:
            self.log(u"  The fragment list does not have fragments with zero length")

    def _process_long_nonspeech(self, ns_min, ns_string, real_wave_mfcc):
        self.log(u"Called _process_long_nonspeech")
        if ns_min is not None:
            self.log(u"Processing long nonspeech intervals requested")
            self.log(u"  Checking and fixing...")
            tolerance = self.rconf[RuntimeConfiguration.ABA_NONSPEECH_TOLERANCE]
            self.log([u"    Tolerance: %.3f", tolerance])
            long_nonspeech_intervals = [i for i in real_wave_mfcc.intervals(speech=False, time=True) if i.length >= ns_min]
            pairs = self.smflist.fragments_ending_inside_nonspeech_intervals(long_nonspeech_intervals, tolerance)
            # ignore HEAD and TAIL
            min_index = 1
            max_index = len(self.smflist) - 1
            pairs = [(n, i) for (n, i) in pairs if (i >= min_index) and (i < max_index)]
            self.smflist.inject_long_nonspeech_fragments(pairs, ns_string)
            self.log(u"  Checking and fixing... done")
        else:
            self.log(u"Processing long nonspeech intervals not requested: returning")

    def _smooth_fragment_list(self, real_wave_mfcc_audio_length, ns_string):
        """
        Remove NONSPEECH fragments from list if needed,
        and set HEAD/TAIL begin/end.
        """
        self.log(u"Called _smooth_fragment_list")
        self.smflist[0].begin = TimeValue("0.000")
        self.smflist[-1].end = real_wave_mfcc_audio_length
        if ns_string in [None, gc.PPV_TASK_ADJUST_BOUNDARY_NONSPEECH_REMOVE]:
            self.log(u"Remove all NONSPEECH fragments")
            self.smflist.remove_nonspeech_fragments(zero_length_only=False)
        else:
            self.log(u"Remove NONSPEECH fragments with zero length only")
            self.smflist.remove_nonspeech_fragments(zero_length_only=True)

    # #####################################################
    # ADJUST FUNCTIONS
    # #####################################################

    def _adjust_auto(self, real_wave_mfcc, algo_parameters):
        """
        AUTO (do not modify)
        """
        self.log(u"Called _adjust_auto")
        self.log(u"Nothing to do, return unchanged")

    def _adjust_offset(self, real_wave_mfcc, algo_parameters):
        """
        OFFSET
        """
        self.log(u"Called _adjust_offset")
        self._apply_offset(offset=algo_parameters[0])

    def _adjust_percent(self, real_wave_mfcc, algo_parameters):
        """
        PERCENT
        """
        def new_time(nsi):
            """
            The new boundary time value is ``percent``
            of the nonspeech interval ``nsi``.
            """
            percent = Decimal(algo_parameters[0])
            return nsi.percent_value(percent)
        self.log(u"Called _adjust_percent")
        self._adjust_on_nonspeech(real_wave_mfcc, new_time)

    def _adjust_aftercurrent(self, real_wave_mfcc, algo_parameters):
        """
        AFTERCURRENT
        """
        def new_time(nsi):
            """
            The new boundary time value is ``delay`` after
            the begin of the nonspeech interval ``nsi``.
            If ``nsi`` has length less than ``delay``,
            set the new boundary time to the end of ``nsi``.
            """
            delay = max(algo_parameters[0], TimeValue("0.000"))
            return min(nsi.begin + delay, nsi.end)
        self.log(u"Called _adjust_aftercurrent")
        self._adjust_on_nonspeech(real_wave_mfcc, new_time)

    def _adjust_beforenext(self, real_wave_mfcc, algo_parameters):
        """
        BEFORENEXT
        """
        def new_time(nsi):
            """
            The new boundary time value is ``delay`` before
            the end of the nonspeech interval ``nsi``.
            If ``nsi`` has length less than ``delay``,
            set the new boundary time to the begin of ``nsi``.
            """
            delay = max(algo_parameters[0], TimeValue("0.000"))
            return max(nsi.end - delay, nsi.begin)
        self.log(u"Called _adjust_beforenext")
        self._adjust_on_nonspeech(real_wave_mfcc, new_time)

    def _adjust_rate(self, real_wave_mfcc, algo_parameters):
        """
        RATE
        """
        self.log(u"Called _adjust_rate")
        self._apply_rate(max_rate=algo_parameters[0], aggressive=False)

    def _adjust_rate_aggressive(self, real_wave_mfcc, algo_parameters):
        """
        RATEAGGRESSIVE
        """
        self.log(u"Called _adjust_rate_aggressive")
        self._apply_rate(max_rate=algo_parameters[0], aggressive=True)

    # #####################################################
    # HELPER FUNCTIONS
    # #####################################################

    def _apply_offset(self, offset):
        """
        Apply the given offset (negative, zero, or positive)
        to all time intervals.

        :param offset: the offset, in seconds
        :type  offset: :class:`~aeneas.exacttiming.TimeValue`
        """
        if not isinstance(offset, TimeValue):
            self.log_exc(u"offset is not an instance of TimeValue", None, True, TypeError)
        self.log([u"Applying offset %s", offset])
        self.smflist.offset(offset)

    def _adjust_on_nonspeech(self, real_wave_mfcc, adjust_function):
        """
        Apply the adjust function to each boundary point
        falling inside (extrema included) of a nonspeech interval.

        The adjust function is not applied to a boundary index
        if there are two or more boundary indices falling
        inside the same nonspeech interval.

        The adjust function is not applied to the last boundary index
        to avoid anticipating the end of the audio file.

        The ``adjust function`` takes
        the nonspeech interval as its only argument.
        """
        self.log(u"Called _adjust_on_nonspeech")
        self.log(u"  Getting nonspeech intervals...")
        nonspeech_intervals = real_wave_mfcc.intervals(speech=False, time=True)
        self.log(u"  Getting nonspeech intervals... done")

        self.log(u"  First pass: find pairs of adjacent fragments transitioning inside nonspeech")
        tolerance = self.rconf[RuntimeConfiguration.ABA_NONSPEECH_TOLERANCE]
        self.log([u"    Tolerance: %.3f", tolerance])
        pairs = self.smflist.fragments_ending_inside_nonspeech_intervals(nonspeech_intervals, tolerance)
        self.log(u"  First pass: done")

        self.log(u"  Second pass: move end point of good pairs")
        for nsi, frag_index, in pairs:
            new_value = adjust_function(nsi)
            self.log([u"    Current interval: %s", self.smflist[frag_index].interval])
            self.log([u"    New value:        %.3f", new_value])
            self.smflist.move_transition_point(frag_index, new_value)
            self.log([u"    New interval:     %s", self.smflist[frag_index].interval])
            self.log(u"")
        self.log(u"  Second pass: done")

    def _apply_rate(self, max_rate, aggressive=False):
        """
        Try to adjust the rate (characters/second)
        of the fragments of the list,
        so that it does not exceed the given ``max_rate``.

        This is done by testing whether some slack
        can be borrowed from the fragment before
        the faster current one.

        If ``aggressive`` is ``True``,
        the slack might be retrieved from the fragment after
        the faster current one,
        if the previous fragment could not contribute enough slack.
        """
        self.log(u"Called _apply_rate")
        self.log([u"  Aggressive: %s", aggressive])
        self.log([u"  Max rate:   %.3f", max_rate])
        regular_fragments = list(self.smflist.regular_fragments)
        if len(regular_fragments) <= 1:
            self.log(u"  The list contains at most one regular fragment, returning")
            return
        faster_fragments = [(i, f) for i, f in regular_fragments if (f.rate is not None) and (f.rate >= max_rate + Decimal("0.001"))]
        if len(faster_fragments) == 0:
            self.log(u"  No regular fragment faster than max rate, returning")
            return
        self.log_warn(u"  Some fragments have rate faster than max rate:")
        self.log([u"  %s", [i for i, f in faster_fragments]])
        self.log(u"Fixing rate for faster fragments...")
        for frag_index, fragment in faster_fragments:
            self.smflist.fix_fragment_rate(frag_index, max_rate, aggressive=aggressive)
        self.log(u"Fixing rate for faster fragments... done")
        faster_fragments = [(i, f) for i, f in regular_fragments if (f.rate is not None) and (f.rate >= max_rate + Decimal("0.001"))]
        if len(faster_fragments) > 0:
            self.log_warn(u"  Some fragments still have rate faster than max rate:")
            self.log([u"  %s", [i for i, f in faster_fragments]])



#!/usr/bin/env python
# coding=utf-8

# aeneas is a Python/C library and a set of tools
# to automagically synchronize audio and text (aka forced alignment)
#
# Copyright (C) 2012-2013, Alberto Pettarin (www.albertopettarin.it)
# Copyright (C) 2013-2015, ReadBeyond Srl   (www.readbeyond.it)
# Copyright (C) 2015-2017, Alberto Pettarin (www.albertopettarin.it)
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
Metadata for setting the aeneas package up
"""

import io

__author__ = "Alberto Pettarin"
__email__ = "aeneas@readbeyond.it"
__copyright__ = """
    Copyright 2012-2013, Alberto Pettarin (www.albertopettarin.it)
    Copyright 2013-2015, ReadBeyond Srl   (www.readbeyond.it)
    Copyright 2015-2017, Alberto Pettarin (www.albertopettarin.it)
"""
__license__ = "GNU AGPL 3"
__status__ = "Production"
__version__ = "1.7.3"


##############################################################################
#
# you might need to edit the information below this line
#
##############################################################################

# package version
# NOTE: generate a new one for each PyPI upload, otherwise it will fail
PKG_VERSION = "1.7.3.0"

# required packages to install
# NOTE: always use exact version numbers
# NOTE: this list should be the same as requirements.txt
PKG_INSTALL_REQUIRES = [
    "BeautifulSoup4>=4.5.1",
    "lxml>=3.6.0",
    "numpy>=1.9"
]

# required packages to install extra tools
PKG_EXTRAS_REQUIRE = {
    "full": [
        "boto3>=1.4.2",
        "Pillow>=3.1.1",
        "requests>=2.9.1",
        "tgt>=1.4.2",
        "youtube-dl>=2016.9.27",
    ],
    "nopillow": [
        "boto3>=1.4.2",
        "requests>=2.9.1",
        "tgt>=1.4.2",
        "youtube-dl>=2016.9.27",
    ],
    "boto3": [
        "boto3>=1.4.2",
    ],
    "pillow": [
        "Pillow>=3.1.1",
    ],
    "requests": [
        "requests>=2.9.1",
    ],
    "tgt": [
        "tgt>=1.4.2",
    ]
}

# packages to be distributed
# NOTE: not including the aeneas.test package to keep the size small
PKG_PACKAGES = [
    "aeneas",
    "aeneas.cdtw",
    "aeneas.cew",
    "aeneas.cfw",
    "aeneas.cmfcc",
    "aeneas.cwave",
    "aeneas.extra",
    "aeneas.syncmap",
    "aeneas.tools",
    "aeneas.ttswrappers"
]

# data files to be distributed
# NOTE: .py files will be added automatically
PKG_PACKAGE_DATA = {
    "aeneas": [
        "res/*",
        "*.md"
    ],
    "aeneas.cdtw": [
        "*.c",
        "*.h",
        "*.md"
    ],
    "aeneas.cew": [
        "*.c",
        "*.h",
        "*.md",
        "*.dll"
    ],
    "aeneas.cew": [
        "*.cc",
        "*.h",
        "*.md"
    ],
    "aeneas.cmfcc": [
        "*.c",
        "*.h",
        "*.md"
    ],
    "aeneas.cwave": [
        "*.c",
        "*.h",
        "*.md"
    ],
    "aeneas.extra": [
        "*.md"
    ],
    "aeneas.syncmap": [
        "*.md"
    ],
    "aeneas.tools": [
        "res/*",
        "*.md"
    ],
    "aeneas.ttswrappers": [
        "*.md"
    ]
}

# scripts to be installed globally
# on Linux and Mac OS X, use the file without extension
# on Windows, use the file with .py extension
PKG_SCRIPTS = [
    "bin/aeneas_check_setup",
    "bin/aeneas_convert_syncmap",
    "bin/aeneas_download",
    "bin/aeneas_execute_job",
    "bin/aeneas_execute_task",
    "bin/aeneas_plot_waveform",
    "bin/aeneas_synthesize_text",
    "bin/aeneas_validate",
]

##############################################################################
#
# do not edit the metadata below this line
#
##############################################################################

# package name
PKG_NAME = "aeneas"

# package author
PKG_AUTHOR = "Alberto Pettarin"

# package author email
PKG_AUTHOR_EMAIL = "alberto@albertopettarin.it"

# package URL
PKG_URL = "https://github.com/readbeyond/aeneas"

# package license
PKG_LICENSE = "GNU Affero General Public License v3 (AGPL v3)"

# human-readable descriptions
PKG_SHORT_DESCRIPTION = "aeneas is a Python/C library and a set of tools to automagically synchronize audio and text (aka forced alignment)"
try:
    PKG_LONG_DESCRIPTION = io.open("README.rst", "r", encoding="utf-8").read()
except:
    PKG_LONG_DESCRIPTION = PKG_SHORT_DESCRIPTION

# PyPI keywords
PKG_KEYWORDS = [
    "AUD",
    "AWS Polly TTS API",
    "CSV",
    "DTW",
    "EAF",
    "ELAN",
    "EPUB 3 Media Overlay",
    "EPUB 3",
    "EPUB",
    "Festival",
    "JSON",
    "MFCC",
    "Mel-frequency cepstral coefficients",
    "Nuance TTS API",
    "ReadBeyond Sync",
    "ReadBeyond",
    "SBV",
    "SMIL",
    "SRT",
    "SSV",
    "SUB",
    "TGT",
    "TSV",
    "TTML",
    "TTS",
    "TextGrid",
    "VTT",
    "XML",
    "aeneas",
    "audio/text alignment",
    "dynamic time warping",
    "eSpeak",
    "eSpeak-ng",
    "espeak",
    "espeak-ng",
    "festival",
    "ffmpeg",
    "ffprobe",
    "forced alignment",
    "media overlay",
    "rb_smil_emulator",
    "speech to text",
    "subtitles",
    "sync",
    "synchronization",
    "text to speech",
    "text2wave",
    "tts",
]

# PyPI classifiers
PKG_CLASSIFIERS = [
    "Development Status :: 5 - Production/Stable",
    "Environment :: Console",
    "Intended Audience :: Developers",
    "Intended Audience :: Education",
    "Intended Audience :: End Users/Desktop",
    "Intended Audience :: Science/Research",
    "License :: OSI Approved :: GNU Affero General Public License v3",
    "Natural Language :: English",
    "Operating System :: MacOS :: MacOS X",
    "Operating System :: Microsoft :: Windows",
    "Operating System :: POSIX :: Linux",
    "Programming Language :: C",
    "Programming Language :: Python",
    "Programming Language :: Python :: 2",
    "Programming Language :: Python :: 2.7",
    "Programming Language :: Python :: 3",
    "Programming Language :: Python :: 3.4",
    "Programming Language :: Python :: 3.5",
    "Topic :: Education",
    "Topic :: Multimedia",
    "Topic :: Multimedia :: Sound/Audio",
    "Topic :: Multimedia :: Sound/Audio :: Analysis",
    "Topic :: Multimedia :: Sound/Audio :: Speech",
    "Topic :: Printing",
    "Topic :: Scientific/Engineering",
    "Topic :: Scientific/Engineering :: Mathematics",
    "Topic :: Software Development :: Libraries :: Python Modules",
    "Topic :: Text Processing",
    "Topic :: Text Processing :: Linguistic",
    "Topic :: Text Processing :: Markup",
    "Topic :: Text Processing :: Markup :: HTML",
    "Topic :: Text Processing :: Markup :: XML",
    "Topic :: Utilities"
]



#!/usr/bin/env python
# coding=utf-8

# aeneas is a Python/C library and a set of tools
# to automagically synchronize audio and text (aka forced alignment)
#
# Copyright (C) 2012-2013, Alberto Pettarin (www.albertopettarin.it)
# Copyright (C) 2013-2015, ReadBeyond Srl   (www.readbeyond.it)
# Copyright (C) 2015-2017, Alberto Pettarin (www.albertopettarin.it)
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
Download a file from a Web source.

Currently, it downloads an audio file from a YouTube video.
"""

from __future__ import absolute_import
from __future__ import print_function
import sys

from aeneas.downloader import Downloader
from aeneas.tools.abstract_cli_program import AbstractCLIProgram
import aeneas.globalfunctions as gf


class DownloadCLI(AbstractCLIProgram):
    """
    Download a file from a Web source.

    Currently, it downloads an audio file from a YouTube video.
    """
    OUTPUT_FILE_M4A = "output/sonnet.m4a"
    OUTPUT_FILE_OGG = "output/sonnet.ogg"
    URL_YOUTUBE = "https://www.youtube.com/watch?v=rU4a7AA8wM0"

    NAME = gf.file_name_without_extension(__file__)

    HELP = {
        "description": u"Download an audio file from a YouTube video.",
        "synopsis": [
            (u"YOUTUBE_URL [OUTPUT_FILE]", True)
        ],
        "examples": [
            u"%s --list" % (URL_YOUTUBE),
            u"%s %s" % (URL_YOUTUBE, OUTPUT_FILE_M4A),
            u"%s %s --format=140" % (URL_YOUTUBE, OUTPUT_FILE_M4A),
            u"%s %s --smallest-audio" % (URL_YOUTUBE, OUTPUT_FILE_OGG),
            u"%s %s --largest-audio" % (URL_YOUTUBE, OUTPUT_FILE_M4A),
        ],
        "options": [
            u"--format=IDX : download audio stream with given format",
            u"--largest-audio : download largest audio stream (default)",
            u"--list : list all available audio streams but do not download",
            u"--smallest-audio : download smallest audio stream",
        ]
    }

    def perform_command(self):
        """
        Perform command and return the appropriate exit code.

        :rtype: int
        """
        if len(self.actual_arguments) < 2:
            return self.print_help()
        source_url = self.actual_arguments[0]
        output_file_path = self.actual_arguments[1]

        download = not self.has_option("--list")
        # largest_audio = True by default or if explicitly given
        if self.has_option("--largest-audio"):
            largest_audio = True
        else:
            largest_audio = not self.has_option("--smallest-audio")
        download_format = self.has_option_with_value("--format")

        try:
            if download:
                self.print_info(u"Downloading audio stream from '%s' ..." % source_url)
                downloader = Downloader(logger=self.logger)
                result = downloader.audio_from_youtube(
                    source_url,
                    download=True,
                    output_file_path=output_file_path,
                    download_format=download_format,
                    largest_audio=largest_audio,
                )
                self.print_info(u"Downloading audio stream from '%s' ... done" % source_url)
                self.print_success(u"Downloaded file '%s'" % result)
            else:
                self.print_info(u"Downloading stream info from '%s' ..." % source_url)
                downloader = Downloader(logger=self.logger)
                result = downloader.audio_from_youtube(
                    source_url,
                    download=False
                )
                self.print_info(u"Downloading stream info from '%s' ... done" % source_url)
                msg = []
                msg.append(u"%s\t%s\t%s\t%s" % ("Format", "Extension", "Bitrate", "Size"))
                for r in result:
                    filesize = gf.human_readable_number(r["filesize"])
                    msg.append(u"%s\t%s\t%s\t%s" % (r["format"], r["ext"], r["abr"], filesize))
                self.print_generic(u"Available audio streams:")
                self.print_generic(u"\n".join(msg))
            return self.NO_ERROR_EXIT_CODE
        except ImportError:
            self.print_no_dependency_error()
        except Exception as exc:
            self.print_error(u"An unexpected error occurred while downloading audio from YouTube:")
            self.print_error(u"%s" % exc)

        return self.ERROR_EXIT_CODE


def main():
    """
    Execute program.
    """
    DownloadCLI().run(arguments=sys.argv)

if __name__ == '__main__':
    main()



#!/usr/bin/env python
# coding=utf-8

# aeneas is a Python/C library and a set of tools
# to automagically synchronize audio and text (aka forced alignment)
#
# Copyright (C) 2012-2013, Alberto Pettarin (www.albertopettarin.it)
# Copyright (C) 2013-2015, ReadBeyond Srl   (www.readbeyond.it)
# Copyright (C) 2015-2017, Alberto Pettarin (www.albertopettarin.it)
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
This module contains the following classes:

* :class:`~aeneas.idsortingalgorithm.IDSortingAlgorithm`,
  enumerating and implementing the available algorithms to sort
  a list of XML ``id`` attributes.

.. warning:: This module is likely to be refactored in a future version
"""

from __future__ import absolute_import
from __future__ import print_function
import re

from aeneas.logger import Loggable


class IDSortingAlgorithm(Loggable):
    """
    Enumeration of the available algorithms to sort
    a list of XML ``id`` attributes.

    :param algorithm: the id sorting algorithm to be used
    :type  algorithm: :class:`~aeneas.idsortingalgorithm.IDSortingAlgorithm`
    :param rconf: a runtime configuration
    :type  rconf: :class:`~aeneas.runtimeconfiguration.RuntimeConfiguration`
    :param logger: the logger object
    :type  logger: :class:`~aeneas.logger.Logger`
    :raises: ValueError: if the value of ``algorithm`` is not allowed
    """

    LEXICOGRAPHIC = "lexicographic"
    """ Lexicographic sorting
    (e.g., ``f020`` before ``f10`` before ``f2``) """

    NUMERIC = "numeric"
    """ Numeric sorting, ignoring any non-digit characters or leading zeroes
    (e.g., ``f2`` (= ``2``) before ``f10`` (= ``10``) before ``f020`` (= ``20``)) """

    UNSORTED = "unsorted"
    """ Do not sort and return the list of identifiers, in their original order
    (e.g., ``f2`` before ``f020`` before ``f10``,
    assuming this was their order in the XML DOM) """

    ALLOWED_VALUES = [LEXICOGRAPHIC, NUMERIC, UNSORTED]
    """ List of all the allowed values """

    TAG = u"IDSortingAlgorithm"

    def __init__(self, algorithm, rconf=None, logger=None):
        if algorithm not in self.ALLOWED_VALUES:
            raise ValueError(u"Algorithm value not allowed")
        super(IDSortingAlgorithm, self).__init__(rconf=rconf, logger=logger)
        self.algorithm = algorithm

    def sort(self, ids):
        """
        Sort the given list of identifiers,
        returning a new (sorted) list.

        :param list ids: the list of identifiers to be sorted
        :rtype: list
        """
        def extract_int(string):
            """
            Extract an integer from the given string.

            :param string string: the identifier string
            :rtype: int
            """
            return int(re.sub(r"[^0-9]", "", string))

        tmp = list(ids)
        if self.algorithm == IDSortingAlgorithm.UNSORTED:
            self.log(u"Sorting using UNSORTED")
        elif self.algorithm == IDSortingAlgorithm.LEXICOGRAPHIC:
            self.log(u"Sorting using LEXICOGRAPHIC")
            tmp = sorted(ids)
        elif self.algorithm == IDSortingAlgorithm.NUMERIC:
            self.log(u"Sorting using NUMERIC")
            tmp = ids
            try:
                tmp = sorted(tmp, key=extract_int)
            except (ValueError, TypeError) as exc:
                self.log_exc(u"Not all id values contain a numeric part. Returning the id list unchanged.", exc, False, None)
        return tmp
